import collections

import numpy as np
from .. import Devutils as dev
from .. import Numputils as nput

__all__ = [
    "canonicalize_internal",
    "is_coordinate_list_like",
    "is_valid_coordinate",
    "permute_internals",
    "find_internal",
    "coordinate_sign",
    "coordinate_indices"
]

def canonicalize_internal(coord, return_sign=False):
    sign = 1
    if len(coord) == 2:
        i, j = coord
        if i == j: return None # faster to just do the cases
        if i > j:
            j, i = i, j
            sign = -1
        coord = (i, j)
    elif len(coord) == 3:
        i, j, k = coord
        if i == j or j == k or i == k: return None
        if i > k:
            i, j, k = k, j, i
            sign = -1
        coord = (i, j, k)
    elif len(coord) == 4:
        i, j, k, l = coord
        if (
                i == j or j == k or i == k
                or i == l or j == l or k == l
        ): return None
        if i > l:
            i, j, k, l = l, k, j, i
            sign = -1
        coord = (i, j, k, l)
    else:
        if len(np.unique(coord)) < len(coord): return None
        if coord[0] > coord[-1]:
            coord = tuple(reversed(coord))
            sign = -1
        else:
            coord = tuple(coord)
    if return_sign:
        return coord, sign
    else:
        return coord

def is_valid_coordinate(coord):
    return (
        len(coord) > 1 and len(coord) < 5
        and all(nput.is_int(c) for c in coord)
    )

def is_coordinate_list_like(clist):
    return dev.is_list_like(clist) and all(
        is_valid_coordinate(c) for c in clist
    )

class InternalsSet:
    def __init__(self, coord_specs:'list[tuple[int]]', prepped_data=None):
        self.specs = coord_specs
        if prepped_data is not None:
            self._indicator, self.coordinate_indices, self.ind_map, self.coord_map = prepped_data
        else:
            self._indicator, self.coordinate_indices, self.ind_map, self.coord_map = self.prep_coords(coord_specs)

    IndicatorMap = collections.namedtuple("IndicatorMap", ['primary', 'child'])
    IndsMap = collections.namedtuple("IndsMap", ['dists', 'angles', 'diheds'])
    InternalsMap = collections.namedtuple("InternalsMap", ['dists', 'angles', 'diheds'])
    @classmethod
    def prep_coords(cls, coord_specs):
        dist_inds = []
        dists = []
        angle_inds = []
        angles = []
        dihed_inds = []
        diheds = []
        indicator = []
        subindicator = []
        atoms = {}

        for i,c in coord_specs:
            c = canonicalize_internal(c)
            atoms.update(c)
            if len(c) == 2:
                indicator.append(0)
                subindicator.append(len(dists))
                dist_inds.append(i)
                dists.append(c)
            elif len(c) == 2:
                indicator.append(1)
                angle_inds.append(i)
                subindicator.append(len(angles))
                angles.append(c)
            elif len(c) == 4:
                indicator.append(2)
                subindicator.append(len(diheds))
                dihed_inds.append(i)
                diheds.append(c)
            else:
                raise ValueError(f"don't know what to do with coord spec {c}")

        return (
            cls.IndicatorMap(np.array(indicator), np.array(subindicator)),
            tuple(sorted(atoms)),
            cls.IndsMap(np.array(dist_inds), np.array(angle_inds), np.array(dihed_inds)),
            cls.InternalsMap(np.array(dists), np.array(angles), np.array(diheds))
        )

    @classmethod
    def _map_dispatch(cls, map, coord):
        if nput.is_int(coord):
            if coord == 0:
                return map.dists
            elif coord == 1:
                return map.angles
            else:
                return map.diheds
        else:
            if len(coord) == 2:
                return map.dists
            elif len(coord) == 3:
                return map.dists
            elif len(coord) == 4:
                return map.diheds
            else:
                raise ValueError(f"don't know what to do with coord spec {coord}")

    def _coord_map_dispatch(self, coord):
        return self._map_dispatch(self.coord_map, coord)
    def _ind_map_dispatch(self, i):
        return self._map_dispatch(self.ind_map, i)
    def find(self, coord):
        return nput.find(self._coord_map_dispatch(coord), coord)

    @classmethod
    def get_coord_from_maps(cls, item, indicator:IndicatorMap, ind_map, coord_map):
        if nput.is_int(item):
            map = indicator.primary[item]
            subloc = indicator.child[item]
            c_map = cls._map_dispatch(coord_map, map)
            return c_map[subloc,]
        else:
            map = indicator.primary[item,]
            uinds = np.unique(map)
            if len(uinds) > 1:
                return [
                    cls.get_coord_from_maps(i, indicator, ind_map, coord_map)
                    for i in item
                ]
            else:
                subloc = indicator.child[item,]
                c_map = cls._map_dispatch(coord_map, uinds[0])
                return c_map[subloc,]

    def __getitem__(self, item):
        return self.get_coord_from_maps(item, self._indicator, self.ind_map, self.coord_map)

    @classmethod
    def _create_coord_list(cls, indicator, inds, vals:InternalsMap):
        #TODO: make this more efficient, just concat the sub
        map = np.argsort(indicator.child)
        full = vals.diheds.tolist() + vals.angles.tolist() + vals.diheds.tolist()
        return [ tuple(full[i]) for i in map ]
    def permute(self, perm, canonicalize=True):
        #TODO: handle padding this
        inv = np.argsort(perm)
        dists = self.coord_map.dists
        if len(dists) > 0:
            dists = inv[dists]
        angles = self.coord_map.angles
        if len(angles) > 0:
            angles = inv[angles]
        diheds = self.coord_map.diheds
        if len(diheds) > 0:
            diheds = inv[diheds]

        cls = type(self)
        int_map = self.InternalsMap(dists, angles, diheds)
        if canonicalize:
            return cls(self._create_coord_list(self._indicator, self.ind_map, int_map))
        else:
            return cls(None, prepped_data=[self._indicator, self.coordinate_indices, self.ind_map, int_map])

def find_internal(coords, coord):
    if isinstance(coords, InternalsSet):
        return coords.find(coord)
    else:
        try:
            idx = coords.index(coord)
        except IndexError:
            idx = None

        if idx is None:
            idx = coords.index(canonicalize_internal(coord))
        return idx

def permute_internals(coords, perm, canonicalize=True):
    if isinstance(coords, InternalsSet):
        return coords.permute(perm, canonicalize=canonicalize)
    else:
        return [
            canonicalize_internal([perm[c] if c < len(perm) else c for c in coord])
                if canonicalize else
            tuple(perm[c] if c < len(perm) else c for c in coord)
            for coord in coords
        ]

def coordinate_sign(old, new, canonicalize=True):
    if len(old) != len(new): return 0
    if len(old) == 2:
        i,j = old
        m,n = new
        if i == n:
            return int(j == m)
        elif i == m:
            return int(i == n)
        else:
            return 0
    elif len(old) == 3:
        i,j,k = old
        m,n,o = new
        if j != n:
            return 0
        elif i == m:
            return int(k == o)
        elif i == o:
            return int(k == m)
        else:
            return 0
    elif len(old) == 4:
        # all pairwise comparisons now too slow
        if canonicalize:
            old = canonicalize_internal(old)
            new = canonicalize_internal(new)

        i,j,k,l = old
        m,n,o,p = new

        if i != m or l != p:
            return 0
        elif j == n:
            return int(k == o)
        elif j == o:
            return -int(k == n)
        else:
            return 0
    else:
        raise ValueError(f"can't compare coordinates {old} and {new}")

def coordinate_indices(coords):
    if isinstance(coords, InternalsSet):
        return coords.coordinate_indices
    else:
        return tuple(sorted(
            {x for c in coords for x in c}
        ))


"""
Tools migrated from `Peeves` for profiling code
"""

__all__ = []
from .Timer import *; from .Timer import __all__ as _all
__all__ += _all
from .Profiler import *; from .Profiler import __all__ as _all
__all__ += _all



from ipywidgets import Widget, DOMWidget, register, widget_serialization, CallbackDispatcher
from traitlets import Unicode, Bool, Instance, Dict, List
from ipywidgets.widgets.trait_types import TypedTuple
from ._frontend import module_name, module_version

@register
class HTMLElement(DOMWidget):
    """
    Represents an HTML element that can be interacted with
    and configured richly, in contrast to the limited interactions
    available in most core Jupyter widgets
    """
    _model_name = Unicode('ActiveHTMLModel').tag(sync=True)
    _model_module = Unicode(module_name).tag(sync=True)
    _model_module_version = Unicode(module_version).tag(sync=True)
    _view_name = Unicode('ActiveHTMLView').tag(sync=True)
    _view_module = Unicode(module_name).tag(sync=True)
    _view_module_version = Unicode(module_version).tag(sync=True)
    tagName = Unicode('div').tag(sync=True)
    classList = List().tag(sync=True)
    styleDict = Dict().tag(sync=True)
    elementAttributes = Dict().tag(sync=True, **widget_serialization)
    innerHTML = Unicode('').tag(sync=True)
    textContent = Unicode('').tag(sync=True)
    children = TypedTuple(trait=Instance(Widget)).tag(sync=True, **widget_serialization)
    id = Unicode('').tag(sync=True)
    value = Unicode('').tag(sync=True)
    exportData = Dict().tag(sync=True, **widget_serialization)
    trackInput = Bool(False).tag(sync=True)
    continuousUpdate = Bool(True).tag(sync=True)
    eventPropertiesDict = Dict().tag(sync=True)
    jsHandlers = Dict().tag(sync=True)
    jsAPI = Instance(Widget, allow_none=True).tag(sync=True, **widget_serialization)
    onevents = Dict().tag(sync=True)
    defaultEventProperties = List(default_value=[
        "bubbles", "cancelable", "composed",
        "target", "timestamp", "type",
        "key", "repeat",
        "button", "buttons",
        "alKey", "shiftKey", "ctrlKey", "metaKey"
    ]).tag(sync=True)
    _debugPrint = Bool(False).tag(sync=True)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._callbacks = CallbackDispatcher()
        self.on_msg(self._handle)
    def _handle(self, _, msg, __):
        self._callbacks(msg)

    def bind_callback(self, callback, remove=False):
        """Register a callback to execute when a DOM event occurs.
        The callback will be called with one argument, an dict whose keys
        depend on the type of event.
        Parameters
        ----------
        remove: bool (optional)
            Set to true to remove the callback from the list of callbacks.
        """
        self._callbacks.register_callback(callback, remove=remove)
    def reset_callbacks(self):
        """Remove any previously defined callback."""
        self._callbacks.callbacks.clear()

    def __repr__(self):
        body = self.children if len(self.children) > 0 else self.innerHTML if len(self.innerHTML) > 0 else self.textContent
        return "{}({}, {!r}, cls={}, style={})".format(
            type(self).__name__,
            'div' if self.tagName == "" else self.tagName,
            body,
            self.classList,
            self.styleDict
        )

    _here = __file__
    @classmethod
    def jupyterlab_install(self, exec_prefix=None, overwrite=False):
        """
        Attempts to do a basic installation for JupterLab
        :return:
        :rtype:
        """
        import sys, shutil, os, tempfile as tf

        prefix = sys.exec_prefix if exec_prefix is None else exec_prefix
        pkg_root = os.path.dirname(os.path.abspath(self._here))
        pkg_name = os.path.basename(pkg_root)
        src = os.path.join(pkg_root, 'labextension')
        target = os.path.join(prefix, "share", "jupyter", "labextensions", pkg_name)
        copied = False
        if overwrite or not os.path.isdir(target):
            copied = True
            if os.path.exists(target):
                with tf.TemporaryDirectory() as new_loc:
                    try:
                        os.remove(new_loc)
                    except:
                        pass
                os.rename(target, new_loc)
            else:
                new_loc = None
            try:
                shutil.copytree(src, target)
            except:
                if new_loc is not None:
                   os.rename(new_loc, target)

        from IPython.core.display import HTML
        if copied:
            return HTML("<h4>Extension installed to {}. You will need to reload the page to get the widgets to display.</h1>".format(target))
    @classmethod
    def jupyternb_install(self, exec_prefix=None, overwrite=False):
        """
        Attempts to do a basic installation for JupterLab
        :return:
        :rtype:
        """
        import sys, shutil, os, tempfile as tf

        prefix = sys.exec_prefix if exec_prefix is None else exec_prefix
        pkg_root = os.path.dirname(os.path.abspath(self._here))
        pkg_name = os.path.basename(pkg_root)
        src = os.path.join(pkg_root, 'nbextension')
        target = os.path.join(prefix, "share", "jupyter", "nbextensions", pkg_name)
        copied = False
        if overwrite or not os.path.isdir(target):
            copied = True
            if os.path.exists(target):
                with tf.TemporaryDirectory() as new_loc:
                    try:
                        os.remove(new_loc)
                    except:
                        pass
                os.rename(target, new_loc)
            else:
                new_loc = None
            try:
                shutil.copytree(src, target)
            except:
                if new_loc is not None:
                   os.rename(new_loc, target)

        from IPython.core.display import HTML
        if copied:
            return HTML("<h4>Extension installed to {}. You will need to reload the page to get the widgets to display.</h1>".format(target))

    def trigger(self, event, content=None, buffers=None):
        if content is None:
            content = {}
        content = content.copy()
        content['handle'] = event
        return self._send({"method": "trigger", "content": content}, buffers=buffers)
    def call(self, method, content=None, buffers=None):
        if content is None:
            content = {}
        content = content.copy()
        content['handle'] = method
        return self._send({"method": "call", "content": content}, buffers=buffers)


"""
Provides classes that are necessary for managing argument signatures
"""

__all__ = [
    "ArgumentType",
    "ArrayType",
    "PointerType",
    "PrimitiveType",
    "RealType",
    "IntType",
    "BoolType",
    "Argument",
    "FunctionSignature"
]

import abc, ctypes, numpy as np, numpy.ctypeslib as npctypes, re

# TODO: need to finish up with the actual type inference & make it
#   so that we can use a `Argument.from_value` constructor
#   Also need to have call_single do the appropriate restructing of its Arguments
#   list so that stuff can cleanly go into the C-level + add some docstring-support

class ArgumentType(metaclass=abc.ABCMeta):
    """
    Defines a general purpose `ArgumentType` so that we can easily manage complicated type specs
    The basic idea is to define a hierarchy of types that can then convert themselves down to
    a `ctypes`-style spec as well as a C++ argument spec so that we can enable `SharedLibraryFunction`
    to use either the basic `ctypes` FFI or a more efficient, but fragile system based off of extension modules.
    This will be explicitly overridden by the `PrimitiveType`, `ArrayType` and `PointerType` subclasses that provide
    the actual useable classes.
    I'd really live to be integrate with what's in the `typing` module to be able to reuse that type-inference machinery
    """

    @property
    @abc.abstractmethod
    def ctypes_type(self):
        raise NotImplementedError()
    @property
    @abc.abstractmethod
    def cpp_type(self):
        raise NotImplementedError()
    @property
    @abc.abstractmethod
    def types(self):
        raise NotImplementedError()
    @property
    @abc.abstractmethod
    def dtypes(self):
        raise NotImplementedError()
    @property
    @abc.abstractmethod
    def typechar(self):
        raise NotImplementedError()
    @abc.abstractmethod
    def isinstance(self, arg):
        raise NotImplementedError()
    @abc.abstractmethod
    def cast(self, arg):
        raise NotImplementedError()
    @abc.abstractmethod
    def c_cast(self, arg):
        raise NotImplementedError()
class PrimitiveType(ArgumentType):
    """
    Defines a general purpose ArgumentType so that we can easily manage complicated type specs
    The basic idea is to define a hierarchy of types that can then convert themselves down to
    a `ctypes`-style spec as well as a C++ argument spec so that we can enable `SharedLibraryFunction`
    to use either the basic `ctypes` FFI or a more efficient, but fragile system based off of extension modules
    """

    typeset = {}
    def __init__(self,
                 name,
                 ctypes_spec,
                 cpp_spec,
                 capi_spec,
                 python_types,
                 numpy_dtypes,
                 serializer,
                 deserializer
                 ):
        """
        :param name: argument name (e.g. 'double')
        :type name: str
        :param ctypes_spec: the ctypes data-type that arguments of this type would be converted to
        :type ctypes_spec:
        :param cpp_spec: the C++ spec for this type (as a string)
        :type cpp_spec: str
        :param capi_spec: the python C-API string for use in `Py_BuildValue`
        :type capi_spec: str
        :param python_types: the python types that this argument maps onto
        :type python_types: Iterable[type]
        :param numpy_dtypes: the numpy dtypes that this argument maps onto
        :type numpy_dtypes: Iterable[np.dtype]
        :param serializer: a serializer for converting this object into a byte-stream
        :type serializer: Callable
        :param deserializer: a deserializer for converting the byte-stream into a C-level object
        :type deserializer: Callable
        """
        self._name = name
        self._ctypes_spec = ctypes_spec
        self._cpp_spec = cpp_spec
        self._capi_spec = capi_spec
        self._types = python_types
        self._dtypes = numpy_dtypes
        self._serializer = serializer
        self._deserializer = deserializer

        self.typeset[self._cpp_spec] = self

    @property
    def name(self):
        return self._name
    @property
    def ctypes_type(self):
        return self._ctypes_spec
    @property
    def cpp_type(self):
        return self._cpp_spec
    @property
    def types(self):
        return self._types
    @property
    def dtypes(self):
        return self._dtypes
    @property
    def typechar(self):
        return self._capi_spec
    def isinstance(self, arg):
        return isinstance(arg, self._types)
    def cast(self, arg):
        return self._types[0](arg)
    def c_cast(self, arg):
        return self.ctypes_type(self.cast(arg))
        # return self._types[0](arg)
    def __repr__(self):
        return "{}({})".format(
            type(self).__name__,
            self.name
        )

class ArrayType(ArgumentType):
    """
    Extends the basic `ArgumentType` spec to handle array types of possibly fixed size.
    To start, we're only adding in proper support for numpy arrays.
    Other flavors might come, but given the use case, it's unlikely.
    """
    def __init__(self, base_type, shape=None, ctypes_spec=None):
        self.base = base_type
        self.shape = shape
        self._ctypes_spec = ctypes_spec

    @property
    def ctypes_type(self):
        if self._ctypes_spec is None:
            self._ctypes_spec = npctypes.ndpointer(self.base.ctypes_type, flags="C_CONTIGUOUS")
        return self._ctypes_spec
    @property
    def cpp_type(self):
        return "*"+self._ctypes_spec
    @property
    def types(self):
        return (np.ndarray,)
    @property
    def dtypes(self):
        return self.base.dtypes
    @property
    def typechar(self):
        return self.base.typechar
    def isinstance(self, arg):
        return isinstance(arg, self.types) and arg.dtype in self.base.dtypes
    def cast(self, arg):
        return np.asanyarray(arg).astype(self.base.dtypes[0])
    def c_cast(self, arg):
        return np.ascontiguousarray(self.cast(arg))
    def __repr__(self):
        return "{}({})".format(
            type(self).__name__,
            self.base
        )

class PointerType(ArgumentType):
    """
    Extends the basic `ArgumentType` spec to handle pointer types
    """
    def __init__(self, base_type):
        """
        :param base_type: The base type we're building off of
        :type base_type: ArgumentType
        """
        self.base = base_type
        self._ctypes_spec = None

    @property
    def ctypes_type(self):
        if self._ctypes_spec is None:
            self._ctypes_spec = ctypes.POINTER(self.base.ctypes_type)
        return self._ctypes_spec

    @property
    def cpp_type(self):
        return "*"+self._ctypes_spec
    @property
    def types(self):
        return self.base.types
    @property
    def dtypes(self):
        return self.base.dtypes
    @property
    def typechar(self):
        return self.base.typechar
    def isinstance(self, arg):
        return self.base.isinstance(arg)
    def cast(self, arg):
        return self.base.cast(arg)
    def c_cast(self, arg):
        # we might need to cast arg first...
        return ctypes.byref(self.base.c_cast(arg))
    def __repr__(self):
        return "{}({})".format(
            type(self).__name__,
            self.base
        )

# this is the block where we just declare a shit ton of types...
# Python types that handle the most common case
VoidType = PrimitiveType(
    "void",
    None,
    "void",
    "void",
    (),
    (),
    None,#serializer
    None#deserializer
)
RealType = PrimitiveType(
    "Real",
    ctypes.c_double,
    "double",
    "d",
    (float,),
    (np.dtype('float64'),),
    None,#serializer
    None#deserializer
)
IntegerType = PrimitiveType(
    "int",
    ctypes.c_int64,
    "int",
    "i",
    (int,),
    (np.dtype('int64'),),
    None,#serializer
    None#deserializer
)
BoolType = PrimitiveType(
    "bool",
    ctypes.c_bool,
    "bool",
    "p",
    (bool,),
    (np.dtype('bool'),),
    None,#serializer
    None#deserializer
)
# C-types with the same names
FloatType = PrimitiveType(
    "float",
    ctypes.c_float,
    "float",
    "f",
    (float,),
    (np.dtype('float32'),),
    None,#serializer
    None#deserializer
)
DoubleType = PrimitiveType(
    "double",
    ctypes.c_double,
    "double",
    "d",
    (float,),
    (np.dtype('float64'),),
    None,#serializer
    None#deserializer
)
IntType = PrimitiveType(
    "int",
    ctypes.c_int32,
    "int",
    "i",
    (int,),
    (np.dtype('int32'),),
    None,#serializer
    None#deserializer
)
LongType = PrimitiveType(
    "long",
    ctypes.c_int64,
    "long",
    "i",
    (int,),
    (np.dtype('int64'),),
    None,#serializer
    None#deserializer
)


class Argument:
    """
    Defines a single Argument for a C-level caller to support default values, etc.
    We use a two-pronged approach where we have a set of ArgumentType serializers/deserializers
    """

    arg_types = [
        VoidType,
        RealType,
        IntType,
        BoolType
    ]

    def __init__(self, name, dtype, default=None):
        """
        :param name: the name of the argument
        :type name: str
        :param dtype: the type of the argument; at some point we'll support type inference...
        :type dtype: ArgumentType
        :param default: the default value for the argument
        :type default:
        """
        self.name = name
        self.dtype = self.infer_type(dtype) #self.infer_type(dtype)
        self.default = default #self.prep_argument(default)
        self._typesets = {}

    @classmethod
    def infer_type(cls, arg):
        """
        Infers the type of an argument

        :param arg:
        :type arg: ArgumentType | str | type | ctypes type
        :return:
        :rtype:
        """

        if isinstance(arg, ArgumentType):
            return arg

        argtype = None
        if isinstance(arg, tuple): # shorthand for pointer types
            argtype = PointerType(cls.infer_type(arg[0]))
        elif isinstance(arg, (list, np.ndarray)):
            argtype = ArrayType(cls.infer_type(arg[0]))
        elif isinstance(arg, str):
            argtype = cls.infer_type_str(arg)
        elif isinstance(arg, type):
            argtype = cls.infer_type_type(arg)
        else:
            for at in cls.arg_types:
                if at.isinstance(arg):
                    argtype = at
                    break

        if argtype is None:
            raise ValueError("can't infer type from {}".format(arg))
        return argtype

    _typesets = None
    _typestrs = None
    @classmethod
    def _prep_typesets(cls):
        if cls._typesets is None or cls._typestrs is None:
            cls._typesets = {}
            cls._typestrs = {}
            for at in cls.arg_types:
                for t in at.types:
                    if t not in cls._typesets:
                        cls._typesets[t] = at
                if at.cpp_type not in cls._typestrs:
                    cls._typestrs[at.cpp_type] = at
    @classmethod
    def infer_type_type(cls, type_key):
        cls._prep_typesets()
        return cls._typesets.get(type_key, None)

    @classmethod
    def infer_type_str(cls, argstr):
        cls._prep_typesets()
        type = None
        if argstr in cls._typesets:
            type = cls._typesets[argstr]
        elif argstr in cls._typestrs:
            type = cls._typestrs[argstr]
        else:
            ptr_pat = "\*({})".format( "|".join(cls._typestrs.keys()))
            match = re.match(ptr_pat, argstr)
            if match is not None:
                typestr = match.group(0)
                type = ArrayType(cls._typestrs[typestr])
        return type

    @classmethod
    def inferred_type_string(cls, arg):
        """
        returns a type string for the inferred type
        """
        raise NotImplementedError("...")

    def prep_value(self, val):
        return self.dtype.c_cast(val)

    def is_pointer(self):
        return isinstance(self.dtype, PointerType)
    def is_array(self):
        return isinstance(self.dtype, ArrayType)
    @property
    def dtypes(self):
        return self.dtype.dtypes
    @property
    def typechar(self):
        return self.dtype.typechar
    @property
    def cpp_signature(self):
        return "{} {}".format(
            self.dtype.cpp_type,
            self.name
        )
    def __repr__(self):
        return "{}('{}', {})".format(
            type(self).__name__,
            self.name,
            self.dtype
        )

class FunctionSignature:
    """
    Defines a function signature for a C-level caller.
    To be used inside `SharedLibraryFunction` and things to manage the core interface.
    """

    def __init__(self, name, *args, defaults=None, return_type=None):
        """
        :param name: the name of the function
        :type name: str
        :param args: the arguments passed to the function
        :type args: Iterable[ArgumentType]
        :param return_type: the return type of the function
        :type return_type: ArgumentType | None
        """
        self.name = name
        if return_type is not None:
            return_type = Argument.infer_type(return_type)
        self._ret_type = return_type
        self._arguments = tuple(self.build_argument(x, i) for i, x in enumerate(args))
        if defaults is None:
            defaults = {}
        self.defaults = defaults

    @classmethod
    def construct(cls, name, defaults=None, return_type=None, **args):
        return FunctionSignature(
            name,
            *(Argument(k, v) for k, v in args.items()),
            defaults=defaults,
            return_type=return_type
        )

    def build_argument(self, argtup, which=None):
        """
        Converts an argument tuple into an Argument object
        :param argtup:
        :type argtup:
        :return:
        :rtype:
        """
        if isinstance(argtup, Argument):
            return argtup
        elif isinstance(argtup, str):
            argtup = ('_argument_{}'.format(which), argtup)
        elif isinstance(argtup, dict):
            if 'default' not in argtup:
                argtup['default'] = None
            argtup = (argtup['name'], argtup['dtype'], argtup['default'])
        if len(argtup) == 2:
            name, dtype = argtup
            default = None
        else:
            name, dtype, default = argtup
        return Argument(name=name, dtype=dtype, default=default)

    @property
    def args(self):
        return self._arguments
    @property
    def return_argtype(self):
        return self._ret_type
    @property
    def return_type(self):
        res = self._ret_type
        if res is not None:
            res = res.ctypes_type
        return res
    @property
    def arg_types(self):
        return [a.dtype.ctypes_type for a in self.args]

    @property
    def cpp_signature(self):
        return "{} {}({})".format(
            "void" if self.return_type is None else self.return_type,
            self.name,
            ", ".join(a.cpp_signature for a in self.args)
        )

    def populate_kwargs(self, args, kwargs, defaults=None):
        kwlist = {}
        for base, val in zip(self.args, args):
            kwlist[base.name] = val
        for k, v in kwargs.items():
            if k in kwlist:
                raise ValueError("got multiple values for argument '{}'".format(k))
            kwlist[k] = v

        if defaults is None:
            defaults = {}

        for a in self.args:
            # resolve any possible default values
            kwlist[a.name] = kwlist.get(
                a.name,
                defaults.get(a.name, self.defaults.get(a.name, a.default))
            )

        return kwlist

    def prep_args(self, args, kwargs, defaults=None):
        if args is not None: # easy way to say no prep needed...
            kwargs = self.populate_kwargs(args, kwargs, defaults=defaults)

        final_args = [
            # coerce to correct type
            a.prep_value(kwargs[a.name])
            for a in self.args
        ]

        return final_args

    def __repr__(self):
        return "{}({}({})->{})".format(
            type(self).__name__,
            self.name,
            ", ".join(repr(a) for a in self.args),
            "Any" if self.return_type is None else self.return_type
        )






import numpy as np
from .CoordinateSystem import BaseCoordinateSystem

__all__ = [
    "CartesianCoordinateSystem",
    "InternalCoordinateSystem",
    "CartesianCoordinateSystem3D",
    "CartesianCoordinates3D",
    "CartesianCoordinates1D",
    "CartesianCoordinates2D",
    "SphericalCoordinateSystem",
    "SphericalCoordinates",
    "ZMatrixCoordinateSystem",
    "ZMatrixCoordinates"
    ]

######################################################################################################
##
##                                   CartesianCoordinateSystem Class
##
######################################################################################################
class CartesianCoordinateSystem(BaseCoordinateSystem):
    """
    Represents Cartesian coordinates generally
    """
    name = "Cartesian"
    def __init__(self, dimension=None, converter_options=None, coordinate_shape=None, **opts):
        """
        :param converter_options: options to be passed through to a `CoordinateSystemConverter`
        :type converter_options: None | dict
        :param dimension: the dimension of the coordinate system
        :type dimension: Iterable[None | int]
        :param opts: other options, if `converter_options` is None, these are used as the `converter_options`
        :type opts:
        """
        if converter_options is None or len(converter_options) == 0:
            converter_options = opts
            opts = {}
        super().__init__(self.name, dimension=dimension, coordinate_shape=coordinate_shape, converter_options=converter_options)
    @classmethod
    def from_state(cls, data, serializer=None):
        return cls(
            dimension=data['dimension'],
            coordinate_shape=data['coordinate_shape'],
            converter_options=data['converter_options']
        )

######################################################################################################
##
##                                   InternalCoordinateSystem Class
##
######################################################################################################
class InternalCoordinateSystem(BaseCoordinateSystem):
    """
    Represents Internal coordinates generally
    """

    name = "Internal"
    def __init__(self, dimension=None, coordinate_shape=None, converter_options=None, **opts):
        """
        :param converter_options: options to be passed through to a `CoordinateSystemConverter`
        :type converter_options: None | dict
        :param coordinate_shape: shape of a single coordinate in this coordiante system
        :type coordinate_shape: Iterable[None | int]
        :param dimension: the dimension of the coordinate system
        :type dimension: Iterable[None | int]
        :param opts: other options, if `converter_options` is None, these are used as the `converter_options`
        :type opts:
        """
        if converter_options is None or len(converter_options) == 0:
            converter_options = opts
            opts = {}
        super().__init__(self.name, dimension=dimension, coordinate_shape=coordinate_shape, converter_options=converter_options)
    @classmethod
    def from_state(cls, data, serializer=None):
        return cls(
            dimension=data['dimension'],
            coordinate_shape=data['coordinate_shape'],
            converter_options=data['converter_options']
        )

######################################################################################################
##
##                                   CartesianCoordinates3D Class
##
######################################################################################################
class CartesianCoordinateSystem1D(CartesianCoordinateSystem):
    """
    Represents Cartesian coordinates in 1D
    """
    name = "Cartesian1D"
    def __init__(self, converter_options=None, dimension=(None, 1), **opts):
        """
        :param converter_options: options to be passed through to a `CoordinateSystemConverter`
        :type converter_options: None | dict
        :param dimension: the dimension of the coordinate system
        :type dimension: Iterable[None | int]
        :param opts: other options, if `converter_options` is None, these are used as the `converter_options`
        :type opts:
        """
        if converter_options is None or len(converter_options) == 0:
            converter_options = opts
            opts = {}
        super().__init__(dimension=dimension, converter_options=converter_options)
CartesianCoordinates1D = CartesianCoordinateSystem1D()
CartesianCoordinates1D.__name__ = "CartesianCoordinates1D"
CartesianCoordinates1D.__doc__ = """
    A concrete instance of `CartesianCoordinateSystem3D`
    """
class CartesianCoordinateSystem2D(CartesianCoordinateSystem):
    """
    Represents Cartesian coordinates in 2D
    """
    name = "Cartesian2D"
    def __init__(self, converter_options=None, dimension=(None, 2), **opts):
        """
        :param converter_options: options to be passed through to a `CoordinateSystemConverter`
        :type converter_options: None | dict
        :param dimension: the dimension of the coordinate system
        :type dimension: Iterable[None | int]
        :param opts: other options, if `converter_options` is None, these are used as the `converter_options`
        :type opts:
        """
        if converter_options is None or len(converter_options) == 0:
            converter_options = opts
            opts = {}
        super().__init__(dimension=dimension, converter_options=converter_options)
CartesianCoordinates2D = CartesianCoordinateSystem2D()
CartesianCoordinates2D.__name__ = "CartesianCoordinates2D"
CartesianCoordinates2D.__doc__ = """
    A concrete instance of `CartesianCoordinateSystem3D`
    """

class CartesianCoordinateSystem3D(CartesianCoordinateSystem):
    """
    Represents Cartesian coordinates in 3D
    """
    name = "Cartesian3D"
    def __init__(self, converter_options=None, dimension=(None, 3), **opts):
        """
        :param converter_options: options to be passed through to a `CoordinateSystemConverter`
        :type converter_options: None | dict
        :param dimension: the dimension of the coordinate system
        :type dimension: Iterable[None | int]
        :param opts: other options, if `converter_options` is None, these are used as the `converter_options`
        :type opts:
        """
        if converter_options is None or len(converter_options) == 0:
            converter_options = opts
            opts = {}
        super().__init__(dimension=dimension, converter_options=converter_options)
CartesianCoordinates3D = CartesianCoordinateSystem3D()
CartesianCoordinates3D.__name__ = "CartesianCoordinates3D"
CartesianCoordinates3D.__doc__ = """
    A concrete instance of `CartesianCoordinateSystem3D`
    """

######################################################################################################
##
##                                   ZMatrixCoordinateSystem Class
##
######################################################################################################
class ZMatrixCoordinateSystem(InternalCoordinateSystem):
    """
    Represents ZMatrix coordinates generally
    """
    name = "ZMatrix"
    def __init__(self,
                 converter_options=None,
                 dimension=(None, None),
                 coordinate_shape=(None, 3),
                 **opts):
        """
        :param converter_options: options to be passed through to a `CoordinateSystemConverter`
        :type converter_options: None | dict
        :param coordinate_shape: shape of a single coordinate in this coordiante system
        :type coordinate_shape: Iterable[None | int]
        :param dimension: the dimension of the coordinate system
        :type dimension: Iterable[None | int]
        :param opts: other options, if `converter_options` is None, these are used as the `converter_options`
        :type opts:
        """
        if converter_options is None or len(converter_options) == 0:
            converter_options = opts
        super().__init__(dimension=dimension, coordinate_shape=coordinate_shape, converter_options=converter_options)
        self.jacobian_prep = self.jacobian_prep_coordinates
    @staticmethod
    def jacobian_prep_coordinates(
                                  coord, displacements, values,
                                  dihedral_cutoff=6
                                  ):
        # target_ndim = len(self.dimension) + len(coord) + 1
        extra_dim = displacements.ndim - values.ndim
        raw_displacement_shape = displacements.shape[:-2]
        analytic_order = extra_dim // 2
        dihedrals = values[..., 2]
        central_point = tuple((x - 1) // 2 for x in raw_displacement_shape) # just need a rough check
        ref = dihedrals[central_point]
        for x in central_point:
            ref = ref[np.newaxis]
        # ref = np.broadcast_to(ref, dihedrals.shape)
        true_diffs = dihedrals - ref
        bad_spots = np.where(abs(true_diffs) > dihedral_cutoff)
        if len(bad_spots) > 0 and len(bad_spots[0]) > 0:
            values = values.copy()
            if analytic_order == 0:
                # ref_vals = ref[bad_spots]
                patch_vals = dihedrals[bad_spots]
                patch_signs = np.sign(patch_vals)
                # if we have a negative 2pi, we have a negative displaced val and positive start val
                # in this case we want to do np.pi + (displaced_val - np.pi) = 2pi + displaced_val
                # if we have a positive 2pi, we have a positive displaced_val and negative start val
                # in this case we want to do -np.pi + (displaced_val - np.pi) = -2pi + displaced_val
                fix_spots = bad_spots + (np.full(len(bad_spots[0]), 2),)
                values[fix_spots] = patch_vals + (-patch_signs)*2*np.pi
            elif analytic_order == 1:
                raise NotImplementedError('correcting periodicity wraparound not handled for analytic derivative order {}'.format(analytic_order))
            else:
                raise NotImplementedError('correcting periodicity wraparound not handled for analytic derivative order {}'.format(analytic_order))

        # we will want to make sure all angles and dihedrals stay within a range of eachother...
        return displacements, values

    def to_state(self, serializer=None):
        base = super().to_state(serializer=serializer)
        converter_options = base.get('converter_options', {})
        if converter_options.get('jacobian_prep') is ZMatrixCoordinateSystem.jacobian_prep_coordinates:
            converter_options = converter_options.copy()
            del converter_options['jacobian_prep']
            base['converter_options'] = converter_options
        return base

    @classmethod
    def canonicalize_order_list(self, ncoords, order_list):
        """
        Normalizes the way the ZMatrix coordinates are built out

        :param ncoords:
        :type ncoords:
        :param order_list: the basic ordering to apply for the
        :type order_list: iterable or None
        :return:
        :rtype: iterator of int triples
        """
        if order_list is None:
            normalized_list = np.array((
                np.arange(ncoords),
                np.arange(-1, ncoords - 1),
                np.arange(-2, ncoords - 2),
                np.arange(-3, ncoords - 3)
            )).T
        else:
            normalized_list = [[]] * len(order_list)
            any3 = None
            def bad_ol():
                raise ValueError((
                    "order list {ol} mixes internal spec forms "
                    "Expected ({n}, {d}, {a}, {t}), ({n}, {d}, {a}, {t}, {f}), ({d}, {a}, {t}), or {n}"
                ).format(
                    ol=order_list,
                    el=el,
                    n="atomNum",
                    d="distAtomNum",
                    a="angleAtomNum",
                    t="dihedAtomNum",
                    f="dihedForm"
                ))
            for i, el in enumerate(order_list):
                if isinstance(el, int):
                    if any3 is None:
                        any3 = True
                    elif not any3:
                        bad_ol()
                    spec = (
                        el,
                        normalized_list[i - 1][0] if i > 0 else -1,
                        normalized_list[i - 2][0] if i > 1 else -1,
                        normalized_list[i - 3][0] if i > 2 else -1
                    )
                else:
                    spec = tuple(el)
                    if len(spec) == 3:
                        if any3 is None:
                            any3 = True
                        elif not any3:
                            bad_ol()
                        spec = (i,) + spec + (
                            normalized_list[i - 1][0] if i > 0 else -1,
                            normalized_list[i - 2][0] if i > 1 else -1,
                            normalized_list[i - 3][0] if i > 2 else -1
                        )
                    elif len(spec) == 4:
                        if any3:
                            bad_ol()
                        any3 = False
                    elif len(spec) == 5:
                        if any3:
                            bad_ol()
                        any3 = False
                    else:
                        raise ValueError(
                            "Z-matrix conversion spec {el} not understood. "
                            "Expected ({n}, {d}, {a}, {t}), ({n}, {d}, {a}, {t}, {f}), ({d}, {a}, {t}), or {n}".format(
                                el=el,
                                n="atomNum",
                                d="distAtomNum",
                                a="angleAtomNum",
                                t="dihedAtomNum",
                                f="dihedForm"
                            ))
                normalized_list[i] = spec
            nlist_len = max(len(x) for x in normalized_list)
            if nlist_len == 5:
                for i,s in enumerate(normalized_list):
                    if len(s) == 4:
                        normalized_list[i] = s + (0,)
        return np.asanyarray(normalized_list, dtype=np.int8)

    @classmethod
    def tile_order_list(self, ol, ncoords):
        nol = len(ol)
        ncol = len(ol[0])
        fsteps = ncoords / nol
        steps = int(fsteps)
        if steps != fsteps:
            raise ValueError(
                "{}: Number of coordinates {} and number of specifed elements {} misaligned".format(
                    type(self),
                    ncoords,
                    nol
                )
            )
        # broadcasts a single order spec to be a multiple order spec
        base_tile = np.broadcast_to(ol[:, :4], (steps, nol, 4))
        shift = np.reshape(np.arange(0, ncoords, nol), (steps, 1, 1))
        ol_tiled = base_tile + shift
        # now we add on extra flags
        if ncol > 4:
            flags_tiled = np.broadcast_to(ol[:, 4:], (steps, nol, ncol-4))
            ol_tiled = np.concatenate([ol_tiled, flags_tiled], axis=-1)
        return np.reshape(ol_tiled, (ncoords, ncol))
ZMatrixCoordinates = ZMatrixCoordinateSystem()
ZMatrixCoordinates.__name__ = "ZMatrixCoordinates"
ZMatrixCoordinates.__doc__ = """
    A concrete instance of `ZMatrixCoordinateSystem`
    """

######################################################################################################
##
##                                   SphericalCoordinateSystem Class
##
######################################################################################################
class SphericalCoordinateSystem(BaseCoordinateSystem):
    """
    Represents Spherical coordinates generally
    """
    name = "SphericalCoordinates"
    def __init__(self, converter_options=None, **opts):
        """
        :param converter_options: options to be passed through to a `CoordinateSystemConverter`
        :type converter_options: None | dict
        :param opts: other options, if `converter_options` is None, these are used as the `converter_options`
        :type opts:
        """
        if converter_options is None or len(converter_options) == 0:
            converter_options = opts
            opts = {}
        super().__init__(self.name, dimension=3, converter_options=converter_options)
SphericalCoordinates = SphericalCoordinateSystem()
SphericalCoordinates.__name__ = "SphericalCoordinates"
SphericalCoordinates.__doc__ = """
    A concrete instance of `SphericalCoordinateSystem`
    """




__all__ = [
    "ASEMolecule"
]

import sys

import numpy as np, io, os
from .. import Numputils as nput

from .ExternalMolecule import ExternalMolecule
from .ChemToolkits import ASEInterface

class ASEMolecule(ExternalMolecule):
    """
    A simple interchange format for ASE molecules
    """

    @property
    def atoms(self):
        return self.mol.symbols
    @property
    def coords(self):
        return self.mol.positions
    @property
    def charges(self):
        return self.mol.charges

    @classmethod
    def from_coords(cls, atoms, coords, charge=None, spin=None, info=None, calculator=None, **etc):
        if calculator is not None and charge is not None:
            if hasattr(calculator, 'set_charge'):
                calculator.set_charge(charge)

        if info is None and charge is not None or spin is not None:
            info = {}
        if charge is not None:
            info['charge'] = charge
        if spin is not None:
            info['spin'] = spin

        return cls(
            ASEInterface.Atoms(
                atoms,
                coords,
                calculator=calculator,
                info=info,
                **etc
            )
        )

    @classmethod
    def from_mol(cls, mol, coord_unit="Angstroms", calculator=None):
        from ..Data import UnitsData

        return cls.from_coords(
            mol.atoms,
            mol.coords * UnitsData.convert(coord_unit, "Angstroms"),
            # bonds=mol.bonds,
            charge=mol.charge,
            calculator=calculator
        )

    # def calculate_gradient(self, geoms=None, force_field_generator=None, force_field_type='mmff'):
    #     if force_field_generator is None:
    #         force_field_generator = self.get_force_field
    #     cur_geom = np.array(self.mol.GetPositions()).reshape(-1, 3)
    #     if geoms is not None:
    #         geoms = np.asanyarray(geoms)
    #         base_shape = geoms.shape[:-2]
    #         geoms = geoms.reshape((-1,) + cur_geom.shape)
    #         vals = np.empty((len(geoms), np.prod(cur_geom.shape, dtype=int)), dtype=float)
    #         try:
    #             for i, g in enumerate(geoms):
    #                 self.mol.SetPositions(g)
    #                 ff = force_field_generator(force_field_type)
    #                 vals[i] = ff.CalcGrad()
    #         finally:
    #             self.mol.SetPositions(cur_geom)
    #         return vals.reshape(base_shape + (-1,))
    #     else:
    #         ff = force_field_generator(force_field_type)
    #         return np.array(ff.CalcGrad()).reshape(-1)

    # def get_calculator(self):
    #     if isinstance(self.calc, MassWeightedCalculator):
    #         return self.calc.copy()
    #     else:
    #         return self.load_class()(self.calc.base_calc)
    def calculate_props(self, props, geoms=None, calc=None, extra_calcs=None):
        from ase.calculators.calculator import all_changes
        if calc is None:
            calc = self.mol.calc
        if geoms is None:
            calc.calculate(self.mol, properties=props, system_changes=all_changes)
            base = {
                k:calc.results[k]
                for k in props
            }
            if extra_calcs is not None:
                updates = extra_calcs(self.mol)
                base.update(updates)
            return base
        else:
            cur_geom = self.mol.positions
            geoms = np.asanyarray(geoms)
            base_shape = geoms.shape[:-2]
            geoms = geoms.reshape((-1,) + cur_geom.shape)
            prop_arrays = {}

            try:
                for i, g in enumerate(geoms):
                    self.mol.positions = g
                    calc.calculate(self.mol, properties=props, system_changes=all_changes)
                    for k in props:
                        res = calc.results[k]
                        if k not in prop_arrays:
                            if hasattr(res, 'shape'):
                                shp = res.shape
                            else:
                                shp = ()
                            prop_arrays[k] = np.empty(
                                (len(geoms),) + shp,
                                dtype=type(res) if not hasattr(res, 'dtype') else res.dtype
                            )
                        prop_arrays[k][i] = res
                    if extra_calcs is not None:
                        updates = extra_calcs(self.mol)
                        for k,res in updates.items():
                            if k not in prop_arrays:
                                prop_arrays[k] = np.empty(
                                    (len(geoms),) + res.shape,
                                    dtype=res.dtype
                                )
                            prop_arrays[k][i] = res
            finally:
                self.mol.positions = cur_geom

            return {
                k:r.reshape(base_shape + r.shape[1:])
                for k,r in prop_arrays.items()
            }

    def calculate_energy(self, geoms=None, order=None, calc=None):
        if calc is None:
            calc = self.mol.calc
        just_eng = order is None
        if just_eng: order = 0
        props = ['energy']
        if order > 0:
            props.append('forces')
        extra_calcs = None
        if order > 1:
            if hasattr(calc, 'get_hessian'):
                extra_calcs = lambda m:{'hessian':calc.get_hessian(m)}
            else:
                raise ValueError("ASE calculators only need to implement forces")
        if order > 2:
            raise ValueError("ASE calculators don't support 3rd derivatives by default")
        res = self.calculate_props(props, geoms=geoms, calc=calc, extra_calcs=extra_calcs)
        if just_eng:
            return res['energy']

        base_ndim = 0 if geoms is None else np.asarray(geoms).ndim - 2
        ncoord = 3 * len(self.masses)

        ret_tup = (res['energy'],)
        if order > 0:
            ret_tup = ret_tup + (
                -res['forces'].reshape(res['forces'].shape[:base_ndim] + (ncoord,)),
            )
        if order > 1:
            ret_tup = ret_tup + (
                res['hessian'].reshape(res['forces'].shape[:base_ndim] + (ncoord, ncoord)),
            )
        return ret_tup


    convergence_criterion = 1e-4
    max_steps = 100
    def optimize_structure(self, geoms=None, calc=None, quiet=True, logfile=None, fmax=None, steps=None, **opts):
        BFGS = ASEInterface.submodule('optimize').BFGS

        if logfile is None:
            if quiet:
                logfile = io.StringIO()
            else:
                logfile = sys.stdout

        if calc is None:
            calc = self.mol.calc
        cur_calc = self.mol.calc
        cur_geom = self.mol.positions
        try:
            self.mol.calc = calc
            if geoms is None:
                opt_rea = BFGS(self.mol, logfile=logfile, **opts)
                if fmax is None:
                    fmax = self.convergence_criterion
                if steps is None:
                    steps = self.max_steps
                opt = opt_rea.run(fmax=fmax, steps=steps)
                opt_coords = self.mol.positions
            else:
                cur_geom = self.mol.positions
                geoms = np.asanyarray(geoms)
                base_shape = geoms.shape[:-2]
                geoms = geoms.reshape((-1,) + cur_geom.shape)
                opt = np.empty((len(geoms),), dtype=object)
                opt_coords = np.empty_like(geoms)

                for i, g in enumerate(geoms):
                    self.mol.positions = g
                    opt_rea = BFGS(self.mol, logfile=logfile, **opts)
                    if fmax is None:
                        fmax = self.convergence_criterion
                    if steps is None:
                        steps = self.max_steps
                    opt[i] = opt_rea.run(fmax=fmax, steps=steps)
                    opt_coords[i] = self.mol.positions
                opt = opt.reshape(base_shape)
                opt_coords = opt_coords.reshape(base_shape + opt_coords.shape[1:])
        finally:
            self.mol.calc = cur_calc
            self.mol.positions = cur_geom

        return opt, opt_coords, {}


"""
Provides scaffolding for creating serializers that dump data to a reloadable format.
Light-weight and unsophisticated, but that's what makes this useful..
"""

import abc, numpy as np, json, io, pickle, os, base64, types, warnings
from collections import OrderedDict

__all__= [
    "PseudoPickler",
    "BaseSerializer",
    "JSONSerializer",
    "NumPySerializer",
    "NDarrayMarshaller",
    "HDF5Serializer",
    "YAMLSerializer",
    "ModuleSerializer"
]

import h5py


class PseudoPickler:
    """
    A simple plugin to work _like_ pickle, in that it should
    hopefully support serializing arbitrary python objects, but which
    doesn't attempt to put stuff down to a single `bytearray`, instead
    supporting objects with `to_state` and `from_state` methods by converting
    them to more primitive serializble types like arrays, strings, numbers,
    etc.
    Falls back to naive pickling when necessary.
    """

    _cache = None
    def __init__(self,
                 allow_pickle=False,
                 protocol=1,
                 b64encode=False
                 ):
        self.allow_pickle=allow_pickle
        self.protocol=protocol
        self.b64encode=b64encode

    _primitive_types = (int, float, bool, str,
                        np.integer, np.floating
                        )
    _list_types = (tuple, list)
    _dict_types = (dict, OrderedDict)
    _importable_types = (type, #types.MethodType, types.ModuleType,
                         bytes, bytearray, memoryview,
                         np.dtype
                         )
    _safe_modules = ["numpy", "multiprocessing"]

    def _sanitize_key(self, key):
        # JSON only supports string/int/bool keys
        if not isinstance(key, (str, int, float, bool)):
            raise ValueError("serialized keys must be primitive types")
        return key
    def _to_state(self, obj, cache):
        """
        Tries to extract state for `obj` by walking through the
        object dict

        :param obj:
        :type obj:
        :param cache: cache of written objects to prevent cycles
        :type cache: set
        :return:
        :rtype:
        """

        if obj is None:
            return None
        elif isinstance(obj, self._primitive_types):
            return obj
        elif isinstance(obj, np.ndarray):
            if obj.dtype == np.dtype(object):
                objid = id(obj)
                if objid in cache:
                    raise ValueError("multiple references to single object not allowed ({} already written)".format(obj))
                cache.add(objid)
                convobj = np.array([self.serialize(v, cache) for v in obj.flatten()])
                obj = convobj.reshape(obj.shape)
                # raise ValueError("don't support arbitrary object NumPy data yet")
            return obj
        elif isinstance(obj, self._list_types):
            return type(obj)(self.serialize(v, cache) for v in obj)
        elif isinstance(obj, self._dict_types):
            return type(obj)((self._sanitize_key(k), self.serialize(v, cache)) for k,v in obj.items())
        elif isinstance(obj, self._importable_types):
            return self._to_importable_state(obj)
        elif self._can_import(obj):
            return self._to_importable_state(obj)
        else:
            objid = id(obj)
            if objid in cache:
                raise ValueError("multiple references to single object not allowed ({} already written)".format(obj))
            cache.add(id(obj))
            try:
                odict = dict(obj.__dict__)
            except AttributeError:
                raise_err = True
            else:
                raise_err = False

            if raise_err:
                raise ValueError("can't psuedopickle object of type {} (not a primitive nor supporting `obj.__dict__`): {}".format(type(obj), obj))

            return self.serialize(odict, cache=cache)

    def _to_importable_state(self, obj):
        try:
            dump = pickle.dumps(obj)
        except:
            raise ValueError("unable to pickle {}; needed to be able to restore state".format(obj))
        if self.b64encode:
            dump = base64.b64encode(dump).decode('ascii')
        return {
            "pseudopickle_protocol": self.protocol,
            "pickle_data": dump
        }
    def _can_import(self, obj):
        # print(type(obj).__module__)
        return (
                type(obj).__module__.split(".")[0] in self._safe_modules
                or isinstance(obj, types.BuiltinFunctionType)
        )

    def to_state(self, obj, cache=None):
        """
        Tries to extract state from `obj`, first through its `to_state`
        interface, but that failing by recursively walking the object
        tree

        :param obj:
        :type obj:
        :return:
        :rtype:
        """

        if hasattr(obj, 'to_state') and not isinstance(obj, type):
            # try:
            return obj.to_state(serializer=self)
            # except TypeError: # usually mean we're pickling the class
            #     pass
        try:
            return self._to_state(obj, cache)
        except ValueError:
            if self.allow_pickle:
                dump = pickle.dumps(obj)
                if self.b64encode:
                    dump = base64.b64encode(dump).decode('ascii')
                return {
                    "pseudopickle_protocol": self.protocol,
                    "pickle_data": dump
                }
            else:
                raise ValueError("couldn't serialize {}".format(obj))

    def serialize(self, obj, cache=None):
        """
        Serializes an object first by checking for a `to_state`
        method, and that missing, by converting to primitive-ish types
        in a recursive strategy if the object passes `is_simple`, otherwise
        falling back to `pickle`

        :param obj: object to be serialized
        :type obj:
        :return: spec for the pseudo-pickled data
        :rtype: dict
        """
        if obj is None:
            return obj

        reset_cache = False
        if cache is None:
            if self._cache is None:
                reset_cache = True
                cache = set()
                type(self)._cache = cache
            else:
                cache = self._cache
        state = self.to_state(obj, cache=cache)
        if isinstance(obj, self._primitive_types + self._list_types + self._dict_types + self._importable_types):
            return state

        state = {
            "pseudopickle_protocol": self.protocol,
            "class": self.to_state(type(obj)),
            "state": state
        }

        if reset_cache:
            type(self)._cache = None

        return state

    # def _from_qualname(self, spec):
    #     woof = importlib.import_module(spec['module_name'])
    #     subnames = spec['qual_name'].split('.')
    #     for a in subnames:
    #         woof = getattr(woof, a)
    #     return woof
    def deserialize(self, spec):
        """
        Deserializes from an object spec, dispatching
        to regular pickle where necessary

        :param object:
        :type object:
        :return:
        :rtype:
        """

        if isinstance(spec, dict) and 'pseudopickle_protocol' in spec:
            if 'class' in spec:
                cls = self.deserialize(spec['class'])
                if hasattr(cls, 'from_state'):
                    return cls.from_state(spec['state'], serializer=self)
                else:
                    return {"class": cls, "state": spec['state']}
            elif 'pickle_data' in spec:
                byte_stream = spec['pickle_data']
                if not isinstance(byte_stream, bytes):
                    if isinstance(byte_stream, np.ndarray):
                        byte_stream = byte_stream.tolist()
                        # corner case that I don't think should ever happen in the
                        # use cases I've worked with...
                        if isinstance(byte_stream, list):
                            byte_stream = bytes(byte_stream)
                    if isinstance(byte_stream, str):
                        byte_stream = byte_stream.encode('ascii')
                    byte_stream = base64.b64decode(byte_stream)
                return pickle.loads(byte_stream)
            else:
                raise NotImplementedError("don't know how to pseudo-unpickle from {}".format(spec))
        elif spec is None or isinstance(spec, self._primitive_types):
            return spec
        elif isinstance(spec, self._list_types):
            return type(spec)(self.deserialize(v) for v in spec)
        elif isinstance(spec, self._dict_types):
            return type(spec)((k, self.deserialize(v)) for k,v in spec.items())
        else:
            return spec

class ConvertedData:
    """
    Wrapper class for holding serialized data so we can be sure it's clean
    """
    def __init__(self, data, serializer):
        self.data = data
        self.serializer = serializer

class BaseSerializer(metaclass=abc.ABCMeta):
    """
    Serializer base class to define the interface
    """

    default_extension="" # mostly useful later

    @abc.abstractmethod
    def convert(self, data):
        """
        Converts data into a serializable format
        :param data:
        :type data:
        :return:
        :rtype:
        """
        raise NotImplementedError("BaseSerializer is an abstract class")
    @abc.abstractmethod
    def deconvert(self, data):
        """
        Converts data from a serialized format into a python format
        :param data:
        :type data:
        :return:
        :rtype:
        """
        raise NotImplementedError("BaseSerializer is an abstract class")
    @abc.abstractmethod
    def serialize(self, file, data, **kwargs):
        """
        Writes the data
        :param file:
        :type file:
        :param data:
        :type data:
        :return:
        :rtype:
        """
        raise NotImplementedError("BaseSerializer is an abstract class")
    @abc.abstractmethod
    def deserialize(self, file, **kwargs):
        """
        Loads data from a file
        :param file:
        :type file:
        :return:
        :rtype:
        """
        raise NotImplementedError("BaseSerializer is an abstract class")

class JSONSerializer(BaseSerializer):
    """
    A serializer that makes dumping data to JSON simpler
    """
    default_extension = ".json"
    class BaseEncoder(json.JSONEncoder):
        def __init__(self, *args, pseudopickler=None, allow_pickle=True, **kwargs):
            super().__init__(*args, **kwargs)
            if pseudopickler is None:
                pseudopickler = PseudoPickler(b64encode=True)
            self.allow_pickle=allow_pickle
            self.pickler = pseudopickler
        def default(self, obj):
            if isinstance(obj, np.ndarray):
                return obj.tolist()
            elif isinstance(obj, (np.integer,)):
                return int(obj)
            elif isinstance(obj, (np.floating,)):
                return float(obj)
            else:
                if self.allow_pickle:
                    try:
                        stream = json.JSONEncoder.default(self, obj)
                    except TypeError:
                        stream = self.pickler.serialize(obj)
                    return stream
                else:
                    return json.JSONEncoder.default(self, obj)

    def __init__(self, encoder=None, allow_pickle=True, pseudopickler=None):
        if pseudopickler is None:
            pseudopickler = PseudoPickler(b64encode=True)
        self.pseudopickler = pseudopickler
        self.allow_pickle = allow_pickle
        if encoder is None:
            self.encoder = self.BaseEncoder(pseudopickler=pseudopickler, allow_pickle=allow_pickle)
    def convert(self, data):
        return ConvertedData(self.encoder.encode(data), self)
    def deconvert(self, data):
        return data
    def serialize(self, file, data, **kwargs):
        if not isinstance(data, ConvertedData):
            data = self.convert(data)
        file.write(data.data)
    def deserialize(self, file, key=None, **kwargs):
        dat = json.load(file)
        dat = self.deconvert(dat)
        if key is not None:
            if '/' in key:
                key = key.split("/")
                for k in key:
                    dat = dat[k]
            else:
                dat = dat[key]

        if self.allow_pickle:
            dat = self.pseudopickler.deserialize(dat)


        return dat

class YAMLSerializer(BaseSerializer):
    """
    A serializer that makes dumping data to YAML simpler.
    Doesn't support arbitrary python objects since that hasn't seemed like
    a huge need yet...
    """
    default_extension = ".yml"
    def __init__(self):
        # just checks that we do really have YAML support...
        import yaml as api
        self.api = api

    def convert(self, data):
        return ConvertedData(data, self)
    def deconvert(self, data):
        return data
    def serialize(self, file, data, **kwargs):
        if not isinstance(data, ConvertedData):
            data = self.convert(data)
        data = data.data
        self.api.dump(data, file, **kwargs)
    def deserialize(self, file, key=None, **kwargs):
        dat = self.api.unshare(file)
        dat = self.deconvert(dat)
        if key is not None:
            if '/' in key:
                key = key.split("/")
                for k in key:
                    dat = dat[k]
            else:
                return dat[key]
        else:
            return dat

class NDarrayMarshaller:
    """
    Support class for `HDF5Serializer` and other
    NumPy-friendly interfaces that marshalls data
    to/from NumPy arrays
    """

    def __init__(self,
                 base_serializer=None,
                 allow_pickle=True,
                 psuedopickler=None,
                 allow_records=False,
                 all_dicts=False,
                 converters=None
                 ):

        self.parent=base_serializer
        self.allow_pickle = allow_pickle
        if allow_pickle and psuedopickler is None:
            psuedopickler = PseudoPickler(b64encode=True)
        self._seen_cache = None
        self.psuedopickler = psuedopickler
        self.all_dicts = all_dicts
        self.allow_records = allow_records
        self._converter_dispatch = converters

    # we define a converter layer that will coerce everything to NumPy arrays
    atomic_types = PseudoPickler._primitive_types

    @classmethod
    def get_default_converters(self):
        return OrderedDict((
            ((np.ndarray,), lambda data, cls: cls._iterable_to_numpy(data)),
            ('to_state', lambda x, s: s._psuedo_pickle_to_numpy(x)),
            ('asarray', lambda data, cls: cls._iterable_to_numpy(data.asarray())),
            ((type(None),), lambda x, cls: cls._none_to_none(x)),
            (self.atomic_types, lambda x, cls: cls._literal_to_numpy(x)),
            ((dict,), lambda data, cls: cls._dict_to_numpy(data)),
            ((list, tuple), lambda data, cls: cls._iterable_to_numpy(data))
        ))

    @property
    def converter_dispatch(self):
        if self._converter_dispatch is None:
            return self.get_default_converters()
        else:
            return self._converter_dispatch

    @classmethod
    def _literal_to_numpy(cls, data):
        return np.array([data]).reshape(())

    def _dict_to_numpy(self, data):
        return {k: self.convert(v) for k, v in data.items()}

    def _none_to_none(self, data):
        return None

    def _prep_iterable(self, data):
        if isinstance(data, np.ndarray):
            arr = data
        else:
            try:
                x0 = data[0]
            except IndexError:
                arr = np.asanyarray(data)
            else:
                try_numpy = True
                if isinstance(x0, np.ndarray):
                    l = x0.shape
                else:
                    try:
                        l = len(x0)
                    except TypeError:
                        try_numpy = False
                if try_numpy:
                    for x in data:
                        try:
                            iter(x)
                            if isinstance(x0, np.ndarray):
                                try_numpy = x.shape == l
                            else:
                                try_numpy = len(x) == l
                        except (TypeError, AttributeError):
                            try_numpy = False
                        if not try_numpy: break

                if try_numpy:
                    #TODO: need to reset this later
                    try:
                        warning_type = np.VisibleDeprecationWarning
                    except AttributeError:
                        warning_type = DeprecationWarning
                    warnings.filterwarnings('error', category=warning_type)
                    try:
                        arr = np.asanyarray(data)
                    except (ValueError, warning_type):
                        arr = np.empty(shape=(len(data),), dtype=object)
                        for i, v in enumerate(data):
                            arr[i] = v
                else:
                    arr = np.empty(shape=(len(data),), dtype=object)
                    for i, v in enumerate(data):
                        arr[i] = v

        return arr

    def _iterable_to_numpy(self, data):
        # we do an initial pass to check if data is a numpy array
        # or if it needs to be/can conceivably be converted


        arr = self._prep_iterable(data)

        if arr.dtype == np.dtype(object):
            # map iterable into a stack of datasets :|
            records = [self.convert(v) for i, v in enumerate(data)]
            if self.allow_records and all(isinstance(v, np.ndarray) for v in records):
                # create record dtype...
                ds_dtype = [
                    ('item_{}'.format(i), v.dtype, v.shape if v.shape != (0,) else ())
                    for i, v in enumerate(records)
                ]
                ds_arr = np.recarray((1,), dtype=ds_dtype)
                for i, v in enumerate(records):
                    key = 'item_{}'.format(i)
                    if v.shape != (0,):
                        ds_arr[key] = v
                return ds_arr
            elif self.all_dicts:
                return dict(
                    {
                        '_list_item_' + str(i): v
                        for i, v in enumerate(records)
                    },
                    _list_numitems=self.convert(len(records))
                )
            else:
                return records
        else:
            return arr

    class _pickle_cache:
        def __init__(self, parent):
            self.parent = parent
            self.tree = []
            self.parents = set()
        def __call__(self, key):
            self.add(key)
            return self
        def __enter__(self):
            return self
        def __exit__(self, exc_type, exc_val, exc_tb):
            self.pop()
            if len(self.parents) == 0:
                self.parent._seen_cache = None
        def add(self, key):
            if id(key) in self.parents:
                raise RecursionError("conversion on object of type {} hit an infinite recusion: {}".format(type(key), key))
            self.parents.add(id(key))
            self.tree.append(id(key))
        def pop(self):
            self.parents.remove(self.tree.pop())
    def _psuedo_pickle_to_numpy(self, data):
        if self._seen_cache is None:
            self._seen_cache = self._pickle_cache(self)
        with self._seen_cache(data):
            data = self.psuedopickler.serialize(data)
        # return self._convert(data, allow_pickle=False)
        return self.convert(data)

    def convert(self, data, allow_pickle=None):
        """
        Recursively loop through, test data, make sure HDF5 compatible
        :param data:
        :type data:
        :return:
        :rtype:
        """

        if allow_pickle is None:
            allow_pickle = self.allow_pickle
        cur_pickle = self.allow_pickle
        try:
            self.allow_pickle = allow_pickle
            converter = None
            for k, f in self.converter_dispatch.items():
                if isinstance(k, tuple):  # check if we're dispatching based on type
                    if isinstance(data, k):
                        converter = f
                        break
                elif isinstance(k, str):  # check if we're duck typing based on attributes
                    if hasattr(data, k):
                        converter = f
                        break
                elif k(data):  # assume dispatch key is a callable that tells us if data is compatible
                    converter = f
                    break

            if converter is None and allow_pickle:
                converter = self._default_convert

            if converter is None:
                raise TypeError("no registered converter to coerce {} into HDF5 compatible format".format(data))

            woof = converter(data, self)
            return woof

        finally:
            self.allow_pickle = cur_pickle

    @staticmethod
    def _default_convert(x, converter):
        return converter._psuedo_pickle_to_numpy(x)

    def deconvert(self, data):
        """
        Reverses the conversion process
        used to marshall the data

        :param data:
        :type data:
        :return:
        :rtype:
        """

        # we loop through the keys and recursively build up a dict
        if hasattr(data, 'items'):
            res = {}
            for k, v in data.items():
                if self.parent is not None:
                    v = self.parent.deconvert(v)
                res[k] = v
            if '_list_numitems' in res:
                # actually an iterable but with inconsistent shape
                n_items = res['_list_numitems']
                res = [res['_list_item_' + str(i)] for i in range(n_items)]
            elif list(res.keys()) == ['_data']:  # special case for if we just saved a single array to file
                res = res['_data']
        elif not isinstance(data, np.ndarray):
            try:
                iter(data)
            except TypeError:
                res = data
            else:
                res = [self.parent.deconvert(v) if self.parent is not None else v for v in data]
        else:
            res = data

        # for primitive data...
        if isinstance(res, np.ndarray) and res.shape == ():
            dtype_name = str(res.dtype)
            if '|S' in dtype_name:
                # convert back from bytes to unicode...
                try:
                    res = res.astype(dtype=dtype_name.replace('|S', '<U'))
                except:
                    pass
            res = res.tolist()

        if self.allow_pickle:
            res = self.psuedopickler.deserialize(res)

        return res

    def __call__(self, data, allow_pickle=None):
        if allow_pickle is None:
            allow_pickle = self.allow_pickle
        return self.convert(data, allow_pickle=allow_pickle)

class HDF5Serializer(BaseSerializer):
    """
    Defines a serializer that can prep/dump python data to HDF5.
    To minimize complexity, we always use NumPy & Pseudopickle as an interface layer.
    This restricts what we can serialize, but generally in insignificant ways.
    """
    default_extension = ".hdf5"
    def __init__(self, allow_pickle=True, psuedopickler=None, converters=None):
        import h5py as api
        self.api = api
        self.allow_pickle = allow_pickle
        if allow_pickle and psuedopickler is None:
            psuedopickler = PseudoPickler(b64encode=True)
        self.psuedopickler = psuedopickler
        self.marshaller = NDarrayMarshaller(
            self,
            allow_pickle=allow_pickle,
            psuedopickler=psuedopickler,
            all_dicts=True,
            converters=converters
        )

    def convert(self, data):
        """
        Converts data into format that can be serialized easily

        :param data:
        :type data:
        :return:
        :rtype:
        """
        return ConvertedData(self.marshaller(data), self)

    def _create_dataset(self, h5_obj, key, data):
        """
        Mostly exists to be overridden
        :param h5_obj:
        :type h5_obj:
        :param key:
        :type key:
        :param data:
        :type data:
        :return:
        :rtype:
        """

        if data is None:
            return h5_obj.create_dataset(key, data=self.api.Empty("i"))
        else:
            # try:
            return h5_obj.create_dataset(key, data=data)
            # except:
            #     raise Exception(data.dtype, data)

    def _validate_datatype(self, data):
        return (
            data is None
            or isinstance(data, np.ndarray)
        )
    def _prune_existing(self, h5_obj, key):
        og = h5_obj
        if isinstance(key, str):
            key = [key]
        for k in key[:-1]:
            h5_obj = h5_obj[k]
        try:
            del h5_obj[key[-1]]
        except OSError:
            raise IOError("failed to remove key {} from {}".format(key, og))
    def _destroy_and_add(self, h5_obj, key, data):
        self._prune_existing(h5_obj, key)
        if isinstance(key, str):
            key = [key]
        for subk in key[:-1]:
            try:
                h5_obj = h5_obj.create_group(subk)
            except TypeError:
                raise TypeError("can't coerce {}:{} to HDF5 format".format(key, data))
        try:
            ds = self._create_dataset(h5_obj, key[-1], data)
        except TypeError:
            raise TypeError("can't coerce {}:{} to HDF5 format".format(key, data))
    def _write_data(self, h5_obj, key, data):
        """
        Writes a numpy array into a group
        :param h5_group:
        :type h5_group: h5py.Group
        :param key:
        :type key: str
        :param data:
        :type data: np.ndarray
        :return:
        :rtype:
        """
        if not self._validate_datatype(data):
            raise TypeError('trying to write non-numpy data {} to key "{}"'.format(data, key))

        if data is not None:
            # clean up string datatypes
            dtype_name = str(data.dtype)
            if '<U' in dtype_name:
                # If so, convert the array to one with bytes
                data = data.astype(dtype=dtype_name.replace('<U', '|S'))

        if isinstance(key, str):
            key = [key]
        ds = h5_obj
        for i, subk in enumerate(key):
            try:
                ds = ds[subk] #type: h5py.Dataset
            except KeyError as e:
                if e.args[0] in {
                    "Unable to open object (bad flag combination for message)",
                    "Unable to open object (message type not found)",
                }:
                    raise KeyError("failed to load key {} in {}".format(key, ds))
                else:
                    if isinstance(ds, self.api.Dataset):
                        ds = ds.parent
                    # self._destroy_and_add(h5_obj, key, data)
                    sub = ds
                    for subk in key[i:-1]:
                        try:
                            sub = sub.create_group(subk)
                        except TypeError:
                            raise TypeError("can't coerce {}:{} to HDF5 format".format(key, data))
                    ds = self._create_dataset(sub, key[-1], data)
                    break
            except ValueError:
                if isinstance(ds, self.api.Dataset):
                    ds = ds.parent

                self._destroy_and_add(ds, key[i:], data)
                ds = ds[key[-1]]
                break

        if not isinstance(ds, h5py.Dataset):
            raise TypeError("expected {} got {}".format(h5py.Dataset.__name__, type(ds).__name__))
        # else:
        try:
            dt = ds.dtype
        except AttributeError:
            dt = None
            if data is not None:
                raise ValueError(data)
        if (
                data is None
                or dt != data.dtype and dt.names is not None
        ): # record arrays are a pain
            if data is None:
                data = self.api.Empty("i")
            self._destroy_and_add(h5_obj, key, data)
        else:
            try:
                ds[...] = data
            except (TypeError, AttributeError):
                self._destroy_and_add(h5_obj, key, data)
            except:
                raise IOError("failed to write key '{}' to HDF5 dataset {}".format(key, ds))
        # no need to return stuff, since we're just serializing
    def _write_dict(self, h5_obj, data):
        """
        Writes a dict into a group
        :param h5_group:
        :type h5_group: h5py.Group
        :param data:
        :type data: dict
        :return:
        :rtype:
        """
        for k,v in data.items():
            # print(h5_obj, k)
            # we want to either make a new Group or write the array to the key
            if isinstance(v, dict):
                sub_obj = h5_obj
                if isinstance(k, str):
                    k = [k]
                for i,subk in enumerate(k):
                    try:
                        sub_obj = sub_obj[subk]
                    except KeyError:
                        for subk in k[i:]:
                            sub_obj = sub_obj.create_group(subk)
                        break
                self._write_dict(sub_obj, v)
            else:
                self._write_data(h5_obj, k, v)

    def serialize(self, file, data, **kwargs):
        if not isinstance(data, ConvertedData):
            data = self.convert(data)
        if not isinstance(file, (self.api.File, self.api.Group)):
            file = self.api.File(file, "a")
        data = data.data
        with file:
            if isinstance(data, np.ndarray):
                key = "_data"
                self._write_data(file, key, data)
            else:
                self._write_dict(file, data)

    def deconvert(self, data):
        """
        Converts an HDF5 Dataset into a NumPy array or Group into a dict
        :param data:
        :type data:
        :return:
        :rtype:
        """
        if isinstance(data, self.api.Dataset):
            if data.dtype == self.api.Empty("i"):
                res = None
            else:
                res = np.empty(data.shape, dtype=data.dtype)
                if data.shape != (0,):
                    data.read_direct(res)
                    names = res.dtype.names
                    if names is not None and 'item_0' in names:
                        res = [ res['item_'+str(i)][0] for i in range(len(names))]

            res = self.marshaller.deconvert(res)
        else:
            res = self.marshaller.deconvert(data)

        return res

    @classmethod
    def _extract_key(cls, dataset, key):
        if not isinstance(key, str):
            for k in key:
                dataset = dataset[k]
        else:
            dataset = dataset[key]
        return dataset
    def deserialize(self, file, key=None, **kwargs):
        if not isinstance(file, (self.api.File, self.api.Group)):
            if not isinstance(file, str) and hasattr(file, 'name'):
                file = self.api.File(file.name, "a")
            else:
                file = self.api.File(file, "a")
        with file as cache:
            if key is not None:
                cache = self._extract_key(cache, key)
            return self.deconvert(cache)

class NumPySerializer(BaseSerializer):
    """
    A serializer that implements NPZ dumps
    """

    default_extension = ".npz"

    # we define a converter layer that will coerce everything to NumPy arrays
    atomic_types = (str, int, float)
    converter_dispatch = None
    @classmethod
    def get_default_converters(self):
        return OrderedDict((
        ((np.ndarray,), lambda data, cls: data),
        ('asarray', lambda data, cls: data.asarray()),
        (self.atomic_types, lambda x, cls: cls._literal_to_numpy(x)),
        ((dict,), lambda data, cls: cls._dict_to_numpy(data)),
        ((list, tuple), lambda data, cls: cls._iterable_to_numpy(data))
    ))

    @classmethod
    def get_converters(self):
        if self.converter_dispatch is None:
            return self.get_default_converters()
        else:
            return self.converter_dispatch


    @classmethod
    def _literal_to_numpy(cls, data):
        return np.array([data]).reshape(())

    @classmethod
    def _dict_to_numpy(cls, data):
        return {k: cls._convert(v) for k, v in data.items()}

    @classmethod
    def _iterable_to_numpy(cls, data):
        arr = np.array(data)
        if arr.dtype == np.dtype(object):
            # map iterable into a stack of datasets :|
            return dict({'_list_item_' + str(i): cls._convert(v) for i, v in enumerate(data)},
                        _list_numitems=cls._convert(len(data)))
        else:
            return arr

    @classmethod
    def _convert(cls, data):
        """
        Recursively loop through, test data, make sure NumPy compatible
        :param data:
        :type data:
        :return:
        :rtype:
        """
        converter = None
        for k, f in cls.get_converters().items():
            if isinstance(k, tuple):  # check if we're dispatching based on type
                if isinstance(data, k):
                    converter = f
                    break
            elif isinstance(k, str):  # check if we're duck typing based on attributes
                if hasattr(data, k):
                    converter = f
                    break
            elif k(data):  # assume dispatch key is a callable that tells us if data is compatible
                converter = f
                break

        if converter is None:
            raise TypeError("no registered converter to coerce {} into HDF5 compatible format".format(data))

        return converter(data, cls)

    dict_key_sep="::>|<::" # we pick a somewhat goofy separator to avoid conflicts in most cases
    def _flatten_dict(self, d, sep=None):
        """
        :param d:
        :type d: dict
        :param sep:
        :type sep: str | None
        :return:
        :rtype:
        """
        if sep is None:
            sep = self.dict_key_sep
        new = {}
        for k, v in d.items():
            if isinstance(v, dict):
                d2 = self._flatten_dict(v, sep=sep)
                new.update({k+sep+k2:v2 for k2,v2 in d2.items()})
            else:
                new[k] = v
        return new

    def convert(self, data):
        first_pass = self._convert(data)
        if isinstance(first_pass, dict):
            # we need to flatten out nested dictionaries
            # we're hoping that we won't have massively nested structures
            # since those will totally wreck performance
            data = self._flatten_dict(first_pass)
        else:
            data = first_pass
        return ConvertedData(data, self)

    def _deconvert_val(self, data):
        if isinstance(data, dict):
            # we check to make sure we don't have an implicitly encoded mixed-type list
            if '_list_numitems' in data:
                # actually an iterable but with inconsistent shape
                n_items = data['_list_numitems']
                res = [ data['_list_item_' + str(i)] for i in range(n_items) ]
            else:
                res = {k: self._deconvert_val(v) for k,v in data.items()}
        else:
            res = data
        return res

    def deconvert(self, data, sep=None):
        """
        Unflattens nested dictionary structures so that the original data
        can be recovered
        :param data:
        :type data:
        :param sep:
        :type sep: str | None
        :return:
        :rtype:
        """
        if sep is None:
            sep = self.dict_key_sep
        if isinstance(data, np.ndarray):
            return data
        else:
            data = dict(data.items())
            # now we have to _unflatten_ the flattened dicts -_-
            keys = list(data.keys())
            remapping = [(k.split(sep), k) for k in keys]
            new_data = {}
            for k_list, main_key in remapping:
                where_do_i_go = new_data
                for k in k_list[:-1]:
                    if k not in where_do_i_go:
                        where_do_i_go[k] = {}
                        where_do_i_go = where_do_i_go[k]
                    else:
                        where_do_i_go = where_do_i_go[k]
                where_do_i_go[k_list[-1]] = data[main_key]
            return self._deconvert_val(new_data)

    def serialize(self, file, data, **kwargs):
        if not isinstance(data, ConvertedData):
            data = self.convert(data)
        data = data.data
        if isinstance(data, np.ndarray):
            return np.save(file, data)
        else:
            return np.savez(file, **data)
    def deserialize(self, file, key=None, **kwargs):
        dat = np.load(file)
        dat = self.deconvert(dat)
        if isinstance(dat, np.ndarray):
            return dat
        if key is not None:
            if '/' in key:
                key = key.split("/")
                for k in key:
                    dat = dat[k]
            else:
                return dat[key]
        else:
            return dat

class ModuleSerializer(BaseSerializer):
    """
    A somewhat hacky serializer that supports module-based serialization.
    Writes all module parameters to a dict with a given attribute.
    Serialization doesn't support loading arbitrary python code, but deserialization does.
    Use at your own risk.
    """

    default_extension = ".py"

    default_loader = None
    default_attr = "config"
    def __init__(self, attr=None, loader=None):
        self._loader =loader
        self._attr = attr

    @property
    def loader(self):
        if self._loader is None:
            if self.default_loader is None:
                self.default_loader = self._get_loader()
            return self.default_loader
        else:
            return self._loader
    @property
    def attr(self):
        if self._attr is None:
            return self.default_attr
        else:
            return self._attr
    @classmethod
    def _get_loader(cls):
        from McUtils.Extensions import ModuleLoader
        return ModuleLoader(rootpkg="Configs")

    def convert(self, data):
        return ConvertedData(data, self)
    def deconvert(self, data):
        return data
    def serialize(self, file, data, **kwargs):
        if not isinstance(data, ConvertedData):
            data = self.convert(data)
        data = data.data
        jsoner = JSONSerializer()
        with io.StringIO() as fake:
            jsoner.serialize(fake, data)
            serialized = fake.getvalue()
        print(
            "{} = {}".format(self.attr, serialized),
            file=file
        )
    def deserialize(self, file, key=None, **kwargs):
        module = self.loader.load(file)
        dat = self.deconvert(getattr(module, self.attr))
        if key is not None:
            if '/' in key:
                key = key.split("/")
                for k in key:
                    dat = dat[k]
            else:
                return dat[key]
        else:
            return dat


"""
Provides tooling to call/work with a potential at the C++ level
"""

__all__ = []
from .Module import *; from .Module import __all__ as exposed
__all__ += exposed
from .Loader import *; from .Loader import __all__ as exposed
__all__ += exposed
from .DynamicFFILibrary import *; from .DynamicFFILibrary import __all__ as exposed
__all__ += exposed



import ast, abc, copy
import numpy as np
ellipsis = type(...)

__all__ = [
    'Abstract'
]

class ASTUtils:
    """
    Provides utilities for writing easier AST expressions
    by essentially allowing for the creation of an alternate,
    less-annotated tree that can be manipulated naturally using
    python operations and then can record and replay those ops
    """

    @staticmethod
    def ast_var(var):
        if isinstance(var, str):
            var = ast.Name(id=var, ctx=ast.Load())
        return var
    @classmethod
    def ast_attr(cls, obj, attr):
        if isinstance(obj, str):
            obj = cls.ast_var(obj)
        return ast.Attribute(value=cls.astify(obj), attr=attr, ctx=ast.Load())
    @classmethod
    def astify(cls, val):
        if isinstance(val, AbstractExpr):
            val = val.to_ast()
        elif not isinstance(val, ast.AST):
            if isinstance(val, (int, float, str, np.integer, np.floating, ellipsis)):
                val = ast.Constant(value=val, kind=None)
            else:
                val = cls.convert(val).to_ast()
        return val

    @classmethod
    def prep_args(cls, args):
        return [
            a.to_star() if isinstance(a, AbstractStarOrIter) else
            a for a in args
        ]
    @classmethod
    def ast_call(cls, fn, *args, **kwargs):
        return ast.Call(
            func=cls.astify(fn) if not isinstance(fn, str) else cls.ast_var(fn),
            args=[cls.astify(a) for a in cls.prep_args(args)],
            keywords=[
                ast.keyword(arg=None, value=cls.astify(v.var))
                    if isinstance(v, AbstractStarStar) else
                ast.keyword(arg=k, value=cls.astify(v))

                for k, v in kwargs.items()
            ]
        )

    @classmethod
    def ast_arglist(cls, spec):
        if isinstance(spec, AbstractArgSpec):
            return spec.to_ast()

        if isinstance(spec, str):
            spec = (spec,)
        if isinstance(spec[0], str):
            spec = (spec, {})

        vararg = None
        args = []
        for a in spec[0]:
            if isinstance(a, AbstractStarOrIter):
                a = a.to_star()
            if isinstance(a, str):
                args.append(ast.arg(arg=a, annotation=None, type_comment=None))
            elif isinstance(a, AbstractName):
                args.append(ast.arg(arg=a.name, annotation=None, type_comment=None))
            elif isinstance(a, AbstractStar):
                var = a.iterable
                if isinstance(var, AbstractName):
                    var = var.name
                if not isinstance(var, str):
                    raise NotImplementedError("????")
                vararg = ast.arg(arg=var, annotation=None, type_comment=None)
            else:
                args.append(a)

        varkwarg = None
        kwarg_keys = []
        kwarg_vals = []
        for k, v in spec[1].items():
            if isinstance(v, AbstractStarStar):
                var = v.var
                if isinstance(var, AbstractName):
                    var = var.name
                varkwarg = ast.arg(arg=var, annotation=None, type_comment=None)
            elif isinstance(k, str):
                kwarg_keys.append(ast.arg(arg=k, annotation=None, type_comment=None))
                kwarg_vals.append(cls.astify(v))
            else:
                kwarg_keys.append(k)
                kwarg_vals.append(cls.astify(v))

        return ast.arguments(
            posonlyargs=[],
            args=args,
            vararg=vararg,
            kwonlyargs=kwarg_keys,
            kw_defaults=kwarg_vals,
            kwarg=varkwarg,
            defaults=[]
        )

    @classmethod
    def ast_const(cls, value):
        if not isinstance(value, (str, int, float, np.integer, np.floating)):
            value = cls.astify(value)
        return ast.Constant(value=value, kind=None)

    @classmethod
    def ast_iterable(cls, otype, values):
        return otype(
            elts=[cls.astify(v) for v in values],
            ctx=ast.Load()
        )
    @classmethod
    def ast_list(cls, *values):
        return cls.ast_iterable(
            ast.List,
            values
        )
    @classmethod
    def ast_tuple(cls, *values):
        return cls.ast_iterable(
            ast.Tuple,
            values
        )
    @classmethod
    def ast_set(cls, *values):
        return cls.ast_iterable(
            ast.Set,
            values
        )
    @classmethod
    def ast_dict(cls, **key_pairs):
        return ast.Dict(
            keys=[cls.astify(k) for k in key_pairs.keys()],
            values=[cls.astify(k) for k in key_pairs.values()],
        )

    @classmethod
    def ast_comprehension(cls, var, iterable, filter=None):
        return ast.comprehension(
            target=cls.ast_var(var) if isinstance(var, str) else cls.astify(var),
            iter=cls.astify(iterable),
            ifs=[cls.astify(f) for f in ([filter] if not isinstance(filter, (list, tuple)) else filter)],
            is_async=0 # someday maybe I'll add in async but I just don't need it right now
        )
    @classmethod
    def ast_type_comprehension(cls, otype, expr, comprehension, iterable=None, filter=None):
        return otype(
            elt=expr,
            generators=[
                cls.ast_comprehension(comprehension, iterable, filter=filter)
                if iterable is not None else
                comprehension
            ]
        )
    @classmethod
    def ast_generator(cls, expr, comprehension, iterable=None, filter=None):
        return cls.ast_type_comprehension(
            ast.GeneratorExp,
            expr, comprehension,
            iterable=iterable, filter=filter
        )
    @classmethod
    def ast_list_comprehension(cls, expr, comprehension, iterable=None, filter=None):
        return cls.ast_type_comprehension(
            ast.ListComp,
            expr, comprehension,
            iterable=iterable, filter=filter
        )
    @classmethod
    def ast_set_comprehension(cls, expr, comprehension, iterable=None, filter=None):
        return cls.ast_type_comprehension(
            ast.ListComp,
            expr, comprehension,
            iterable=iterable, filter=filter
        )
    @classmethod
    def ast_dict_comprehension(cls, key_expr, value_expr, comprehension, iterable=None, filter=None):
        return ast.DictComp(
            key=cls.astify(key_expr),
            value=cls.astify(value_expr),
            generators=[
                cls.ast_comprehension(comprehension, iterable, filter=filter)
                if iterable is not None else
                comprehension
            ]
        )

    @classmethod
    def ast_lambda(cls, spec, body):
        return ast.Lambda(
            args=cls.ast_arglist(spec) if not isinstance(spec, ast.arguments) else spec,
            body=cls.astify(body)
        )

    class ContextModifier(ast.NodeTransformer):
        def __init__(self, ctx):
            self.ctx = ctx
        def generic_visit(self, node: ast.AST) -> ast.AST:
            if hasattr(node, 'ctx'):
                node = copy.copy(node)
                node.ctx = self.ctx
            return super().generic_visit(node)

    @classmethod
    def ast_assign(cls, targets, value):
        mod = cls.ContextModifier(ast.Store())
        return ast.Assign(targets=[mod.visit(cls.astify(t)) for t in targets], value=cls.astify(value))

    @classmethod
    def ast_boolop(cls, op, left, right, *extra):
        return ast.BoolOp(
            op=op,
            values=[cls.astify(left), cls.astify(right), *(cls.astify(e) for e in extra)]
        )
    @classmethod
    def ast_and(cls, left, right, *extra):
        return cls.ast_boolop(ast.And(), left, right, *extra)
    @classmethod
    def ast_or(cls, left, right, *extra):
        return cls.ast_boolop(ast.Or(), left, right, *extra)

    @classmethod
    def ast_binop(cls, op, left, right):
        return ast.BinOp(
            left=cls.astify(left),
            op=op,
            right=cls.astify(right)
        )
    @classmethod
    def ast_add(cls, left, right):
        return cls.ast_binop(ast.Add(), left, right)
    @classmethod
    def ast_sub(cls, left, right):
        return cls.ast_binop(ast.Sub(), left, right)
    @classmethod
    def ast_mult(cls, left, right):
        return cls.ast_binop(ast.Mult(), left, right)
    @classmethod
    def ast_div(cls, left, right):
        return cls.ast_binop(ast.Div(), left, right)
    @classmethod
    def ast_matmult(cls, left, right):
        return cls.ast_binop(ast.MatMult(), left, right)
    @classmethod
    def ast_floordiv(cls, left, right):
        return cls.ast_binop(ast.FloorDiv(), left, right)
    @classmethod
    def ast_pow(cls, left, right):
        return cls.ast_binop(ast.Pow(), left, right)
    @classmethod
    def ast_mod(cls, left, right):
        return cls.ast_binop(ast.Mod(), left, right)
    @classmethod
    def ast_bitxor(cls, left, right):
        return cls.ast_binop(ast.BitXor(), left, right)
    @classmethod
    def ast_bitand(cls, left, right):
        return cls.ast_binop(ast.BitAnd(), left, right)
    @classmethod
    def ast_bitor(cls, left, right):
        return cls.ast_binop(ast.BitOr(), left, right)

    @classmethod
    def ast_comp(cls, op, left, right, *extra):
        return ast.Compare(
            left=cls.astify(left),
            op=op,
            comparators=[cls.astify(right), *(cls.astify(v) for v in extra)]
        )
    @classmethod
    def ast_eq(cls, left, right, *extra):
        return cls.ast_comp(ast.Eq(), left, right, *extra)
    @classmethod
    def ast_ne(cls, left, right, *extra):
        return cls.ast_comp(ast.NotEq(), left, right, *extra)
    @classmethod
    def ast_lt(cls, left, right, *extra):
        return cls.ast_comp(ast.Lt(), left, right, *extra)
    @classmethod
    def ast_gt(cls, left, right, *extra):
        return cls.ast_comp(ast.Gt(), left, right, *extra)
    @classmethod
    def ast_lte(cls, left, right, *extra):
        return cls.ast_comp(ast.LtE(), left, right, *extra)
    @classmethod
    def ast_gte(cls, left, right, *extra):
        return cls.ast_comp(ast.GtE(), left, right, *extra)
    @classmethod
    def ast_in(cls, left, right, *extra):
        return cls.ast_comp(ast.In(), left, right, *extra)
    @classmethod
    def ast_ni(cls, left, right, *extra):
        return cls.ast_comp(ast.NotIn(), left, right, *extra)
    @classmethod
    def ast_is(cls, left, right, *extra):
        return cls.ast_comp(ast.Is(), left, right, *extra)
    @classmethod
    def ast_isnt(cls, left, right, *extra):
        return cls.ast_comp(ast.IsNot(), left, right, *extra)

    @classmethod
    def ast_unop(cls, op, val):
        return ast.UnaryOp(
            op=op,
            operand=cls.astify(val)
        )
    @classmethod
    def ast_not(cls, val):
        return cls.ast_unop(ast.Not(), val)
    @classmethod
    def ast_inv(cls, val):
        return cls.ast_unop(ast.Invert(), val)
    @classmethod
    def ast_pos(cls, val):
        return cls.ast_unop(ast.UAdd(), val)
    @classmethod
    def ast_neg(cls, val):
        return cls.ast_unop(ast.USub(), val)

    @classmethod
    def ast_star(cls, iterable):
        return ast.Starred(cls.astify(iterable), ctx=ast.Load())

    @classmethod
    def ast_ternary(cls, test, body, orelse):
        return ast.IfExp(
            test=cls.astify(test),
            body=cls.astify(body),
            orelse=cls.astify(orelse)
        )

    @classmethod
    def ast_subscript(cls, value, index):
        return ast.Subscript(
            value=cls.astify(value),
            slice=cls.ast_index(index) if not isinstance(index, slice) else cls.ast_slice(
                index.start,
                index.stop,
                index.step
            ),
            ctx=ast.Load()
        )

    @classmethod
    def ast_index(cls, value):
        return ast.Index(value=cls.astify(value))
    @classmethod
    def ast_slice(cls, start=None, stop=None, step=None):
        return ast.Slice(
            lower=cls.astify(start) if start is not None else start,
            upper=cls.astify(stop) if stop is not None else stop,
            step=cls.astify(step) if step is not None else step
            )

    @classmethod
    def convert(cls, val) -> 'AbstractExpr':
        if isinstance(val, (str, int, float, np.integer, np.floating, ellipsis)):
            val = AbstractConstant(val)
        elif isinstance(val, list):
            val = AbstractList(*val)
        elif isinstance(val, tuple):
            val = AbstractTuple(*val)
        elif isinstance(val, set):
            val = AbstractSet(*val)
        elif isinstance(val, dict):
            val = AbstractDict(**val)
        elif not isinstance(val, AbstractExpr):
            raise ValueError("no rules for converting {}".format(val))
        return val

class AbstractExpr(metaclass=abc.ABCMeta):
    __tag__ = None
    __slots__ = []
    def __hash__(self):
        return hash((type(self), tuple(getattr(self, k) for k in self.__slots__)))

    @abc.abstractmethod
    def to_ast(self)->'ast.AST':
        ...

    def to_eval_expr(self):
        return ast.fix_missing_locations(ast.Expression(body=self.to_ast()))
    def compile(self, namespace=None):
        if namespace is None:
            namespace = {}
        return eval(
            compile(self.to_eval_expr(), '<symbolic expression>', mode='eval'),
            namespace
        )

    def transmogrify(self, converter_dispatch:dict):
        def convert(expr):
            if not isinstance(expr, AbstractExpr):
                conversion_func = converter_dispatch.get(None, None)
                if conversion_func is None:
                    raise ValueError("missing conversion for object {}".format(expr))
                return conversion_func(expr, convert)

            conversion_func = converter_dispatch.get(expr.__tag__, None)
            if conversion_func is None:
                conversion_func = converter_dispatch.get('Expr', None)
                if conversion_func is None:
                    raise ValueError("missing conversion for tag {}".format(expr.__tag__))
            return conversion_func(expr, convert)
        return convert(self)


    def __call__(self, *args, **kwargs):
        return AbstractCall(self, *args, **kwargs)

    def __not__(self):
        return AbstractNot(self)
    def __invert__(self):
        return AbstractInv(self)
    def __pos__(self):
        return AbstractPos(self)
    def __neg__(self):
        return AbstractNeg(self)

    def __add__(self, other):
        return AbstractAdd(self, other)
    def __radd__(self, other):
        return AbstractAdd(other, self)
    def __sub__(self, other):
        return AbstractSub(self, other)
    def __mul__(self, other):
        return AbstractMult(self, other)
    def __rmul__(self, other):
        return AbstractMult(other, self)
    def __pow__(self, other):
        return AbstractPow(self, other)
    def __truediv__(self, other):
        return AbstractDiv(self, other)
    def __floordiv__(self, other):
        return AbstractFloorDiv(self, other)
    def __xor__(self, other):
        return AbstractBitXOr(self, other)
    def __rxor__(self, other):
        return AbstractBitXOr(other, self)
    def __or__(self, other):
        return AbstractBitOr(self, other)
    def __ror__(self, other):
        return AbstractBitOr(other, self)
    def __and__(self, other):
        return AbstractBitAnd(self, other)
    def __rand__(self, other):
        return AbstractBitAnd(other, self)

    def __contains__(self, item):
        return AbstractIn(self, item)

    def __abs__(self):
        return AbstractAbs(self)
    def __floor__(self):
        return AbstractFloor(self)
    def __ceil__(self):
        return AbstractCeil(self)

    def __iter__(self):
        yield AbstractStarOrIter(self)

    def keys(self):
        yield AbstractUnwrapKeys(self)
    def __getitem__(self, item):
        if isinstance(item, AbstractUnwrapKeys):
            return AbstractStarStar(self)
        else:
            return AbstractSubscript(self, item)

    def __getattr__(self, item):
        return AbstractAttribute(self, item)

    def Equals(self, right):
        return AbstractEq(self, right)
    Eq = Equals
    def LessThan(self, right):
        return AbstractLt(self, right)
    Lt = LessThan
    def LessEquals(self, right):
        return AbstractLtE(self, right)
    LtE = LessEquals
    def GreaterThan(self, right):
        return AbstractGt(self, right)
    Gt = GreaterThan
    def GreaterEquals(self, right):
        return AbstractGtE(self, right)
    GtE = GreaterEquals
    def And(self, right):
        return AbstractAnd(self, right)
    def Or(self, right):
        return AbstractOr(self, right)

    def Is(self, other):
        return AbstractIs(self, other)
    def IsNot(self, other):
        return AbstractIsNot(self, other)
    Isnt = IsNot

    def generator(self, expr, var, filter=None):
        return AbstractGenerator(expr, var, self, filter=filter)
    Gen = generator
    def list_comp(self, expr, var, filter=None):
        return AbstractListComp(expr, var, self, filter=filter)
    LC = list_comp
    def set_comp(self, expr, var, filter=None):
        return AbstractSetComp(expr, var, self, filter=filter)
    SC = set_comp
    def dict_comp(self, key_expr, value_expr, var, filter=None):
        return AbstractDictComp(key_expr, value_expr, var, self, filter=filter)
    DC = dict_comp

    def if_test(self, test, else_expr=None):
        return AbstractIfExp(test, self, else_expr)
    def if_self(self, expr, else_expr=None):
        return AbstractIfExp(self, expr, else_expr)

    def __repr__(self):
        return "{}()".format(type(self).__name__)

class AbstractName(AbstractExpr):
    __tag__ = "Name"
    __slots__ = ['name']
    def __init__(self, name):
        self.name = name
    def to_ast(self)->'ast.Name':
        return ASTUtils.ast_var(self.name)
    def __repr__(self):
        return "{}({})".format(type(self).__name__, self.name)

class AbstractConstant(AbstractExpr):
    __tag__ = 'Const'
    __slots__ = ['value']
    def __init__(self, value):
        self.value = value
    def to_ast(self):
        return ASTUtils.ast_const(self.value)
    def __repr__(self):
        return "{}({})".format(type(self).__name__, self.value)

class AbstractCall(AbstractExpr):
    __tag__ = 'Call'
    __slots__ = ['fn', 'args', 'kwargs']
    def __init__(self, fn, *args, **kwargs):
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
    def to_ast(self)->'ast.Call':
        return ASTUtils.ast_call(self.fn,
                             *self.args,
                             **self.kwargs
                             )
    def __repr__(self):
        return "{}({}, {}, {})".format(type(self).__name__, self.fn, self.args, self.kwargs)

class AbstractAttribute(AbstractExpr):
    __tag__ = 'Attr'
    __slots__ = ['obj', 'attr']
    def __init__(self, obj, attr):
        self.obj = obj
        self.attr = attr
    def to_ast(self) -> 'ast.Attribute':
        return ASTUtils.ast_attr(self.obj, self.attr)
    def __repr__(self):
        return "{}({}, {})".format(type(self).__name__, self.obj, self.attr)

class AbstractBinOp(AbstractExpr):
    __slots__ = ['left', 'right']
    def __init__(self, left, right):
        self.left = left
        self.right = right
class AbstractAdd(AbstractBinOp):
    __tag__ = 'Add'
    def to_ast(self) ->'ast.BinOp':
        return ASTUtils.ast_add(self.left, self.right)
class AbstractSub(AbstractBinOp):
    __tag__ = 'Sub'
    def to_ast(self) ->'ast.BinOp':
        return ASTUtils.ast_sub(self.left, self.right)
class AbstractMult(AbstractBinOp):
    __tag__ = 'Mul'
    def to_ast(self) ->'ast.BinOp':
        return ASTUtils.ast_mult(self.left, self.right)
class AbstractDiv(AbstractBinOp):
    __tag__ = 'Div'
    def to_ast(self) ->'ast.BinOp':
        return ASTUtils.ast_div(self.left, self.right)
class AbstractMatMult(AbstractBinOp):
    __tag__ = 'MatMul'
    def to_ast(self) ->'ast.BinOp':
        return ASTUtils.ast_matmult(self.left, self.right)
class AbstractFloorDiv(AbstractBinOp):
    __tag__ = 'FloorDiv'
    def to_ast(self) ->'ast.BinOp':
        return ASTUtils.ast_floordiv(self.left, self.right)
class AbstractPow(AbstractBinOp):
    __tag__ = 'Pow'
    def to_ast(self) ->'ast.BinOp':
        return ASTUtils.ast_pow(self.left, self.right)
class AbstractMod(AbstractBinOp):
    __tag__ = 'Mod'
    def to_ast(self) ->'ast.BinOp':
        return ASTUtils.ast_mod(self.left, self.right)
class AbstractBitXOr(AbstractBinOp):
    __tag__ = 'BitXOr'
    def to_ast(self) ->'ast.BinOp':
        return ASTUtils.ast_bitxor(self.left, self.right)
class AbstractBitOr(AbstractBinOp):
    __tag__ = 'BitOr'
    def to_ast(self) ->'ast.BinOp':
        return ASTUtils.ast_bitor(self.left, self.right)
class AbstractBitAnd(AbstractBinOp):
    __tag__ = 'BitAnd'
    def to_ast(self) ->'ast.BinOp':
        return ASTUtils.ast_bitand(self.left, self.right)

class AbstractUnOp(AbstractExpr):
    __slots__ = ['operand']
    def __init__(self, operand):
        self.operand = operand
class AbstractNot(AbstractUnOp):
    __tag__ = 'Not'
    def to_ast(self) ->'ast.UnaryOp':
        return ASTUtils.ast_not(self.operand)
class AbstractInv(AbstractUnOp):
    __tag__ = 'Inv'
    def to_ast(self) ->'ast.UnaryOp':
        return ASTUtils.ast_inv(self.operand)
class AbstractPos(AbstractUnOp):
    __tag__ = 'Pos'
    def to_ast(self) ->'ast.UnaryOp':
        return ASTUtils.ast_pos(self.operand)
class AbstractNeg(AbstractUnOp):
    __tag__ = 'Neg'
    def to_ast(self) ->'ast.UnaryOp':
        return ASTUtils.ast_neg(self.operand)

class AbstractComp(AbstractExpr):
    __slots__ = ['left', 'right', 'extra']
    def __init__(self, left, right, *extra):
        self.left = left
        self.right = right
        self.extra = extra
class AbstractEq(AbstractComp):
    __tag__ = 'Equals'
    def to_ast(self) ->'ast.Compare':
        return ASTUtils.ast_eq(self.left, self.right, *self.extra)
class AbstractNotEq(AbstractComp):
    __tag__ = 'NotEquals'
    def to_ast(self) ->'ast.Compare':
        return ASTUtils.ast_ne(self.left, self.right, *self.extra)
class AbstractLt(AbstractComp):
    __tag__ = 'Less'
    def to_ast(self) ->'ast.Compare':
        return ASTUtils.ast_lt(self.left, self.right, *self.extra)
class AbstractGt(AbstractComp):
    __tag__ = 'Greater'
    def to_ast(self) ->'ast.Compare':
        return ASTUtils.ast_gt(self.left, self.right, *self.extra)
class AbstractLtE(AbstractComp):
    __tag__ = 'LessEquals'
    def to_ast(self) ->'ast.Compare':
        return ASTUtils.ast_lte(self.left, self.right, *self.extra)
class AbstractGtE(AbstractComp):
    __tag__ = 'GreaterEquals'
    def to_ast(self) ->'ast.Compare':
        return ASTUtils.ast_gte(self.left, self.right, *self.extra)
class AbstractIn(AbstractComp):
    __tag__ = 'In'
    def to_ast(self) ->'ast.Compare':
        return ASTUtils.ast_in(self.left, self.right, *self.extra)
class AbstractNotIn(AbstractComp):
    __tag__ = 'NotIn'
    def to_ast(self) ->'ast.Compare':
        return ASTUtils.ast_ni(self.left, self.right, *self.extra)
class AbstractIs(AbstractComp):
    __tag__ = 'Is'
    def to_ast(self) ->'ast.Compare':
        return ASTUtils.ast_is(self.left, self.right, *self.extra)
class AbstractIsNot(AbstractComp):
    __tag__ = 'IsNot'
    def to_ast(self) ->'ast.Compare':
        return ASTUtils.ast_isnt(self.left, self.right, *self.extra)

class AbstractSubscript(AbstractExpr):
    __tag__ = 'Subscript'
    __slots__ = ['obj', 'index']
    def __init__(self, obj, index):
        self.obj, self.index = obj, index
    def to_ast(self):
        return ASTUtils.ast_subscript(self.obj, self.index)

class AbstractBoolOp(AbstractExpr):
    __slots__ = ['left', 'right', 'extra']
    def __init__(self, left, right, *extra):
        self.left = left
        self.right = right
        self.extra = extra
class AbstractAnd(AbstractComp):
    __tag__ = 'And'
    def to_ast(self) ->'ast.BoolOp':
        return ASTUtils.ast_and(self.left, self.right, *self.extra)
class AbstractOr(AbstractComp):
    __tag__ = 'Or'
    def to_ast(self) ->'ast.BoolOp':
        return ASTUtils.ast_or(self.left, self.right, *self.extra)


class AbstractImportModule(AbstractCall):
    def __init__(self, mod):
        super().__init__('__import__', mod)

class AbstractAbs(AbstractCall):
    def __init__(self, val):
        super().__init__('abs', val)
class AbstractFloor(AbstractCall):
    def __init__(self, val):
        super().__init__(AbstractImportModule('math').floor, val)
class AbstractCeil(AbstractCall):
    def __init__(self, val):
        super().__init__(AbstractImportModule('math').ceil, val)


class AbstractIter(AbstractCall):
    def __init__(self, val):
        super().__init__('iter', val)
class AbstractStar(AbstractExpr):
    __tag__ = 'Star'
    __slots__ = ['iterable']
    def __init__(self, iterable):
        self.iterable = iterable
    def to_ast(self) ->'ast.Starred':
        return ASTUtils.ast_star(self.iterable)
class AbstractStarOrIter(AbstractExpr):
    __tag__ = 'StarOrIter'
    __slots__ = ['val']
    def __init__(self, val):
        self.val = val
    def to_iter(self):
        return AbstractIter(self.val)
    def to_star(self):
        return AbstractStar(self.val)
    def to_ast(self) ->'ast.AST':
        raise ValueError("type of `AbstractStarOrIter` is indeterminate")

class AbstractUnwrapKeys(str):
    __tag__ = 'UnwrapKeys'
    __slots__ = ['val']
    def __new__(cls, content):
        return str.__new__(cls, "____init_keys" + str(content))
    def __init__(self, val):
        super().__init__()
        self.val = val
    def to_iter(self):
        return AbstractCall(AbstractAttribute(self.val, 'keys'))
    def to_starstar(self):
        return AbstractStarStar(self.val)
    def to_ast(self) -> 'ast.AST':
        raise ValueError("type of `AbstractUnwrapKeys` is indeterminate")
class AbstractStarStar(AbstractExpr):
    __tag__ = 'StarStar'
    __slots__ = ['var']
    def __init__(self, var):
        self.var = var
    def to_ast(self):
        raise ValueError("type of `AbstractStarStar` has not ast equivalent")

class AbstractList(AbstractExpr):
    __tag__ = 'List'
    __slots__ = ['args']
    def __init__(self, *args):
        self.args = args
    def to_ast(self) ->'ast.List':
        return ASTUtils.ast_list(*self.args)
class AbstractTuple(AbstractExpr):
    __tag__ = 'Tuple'
    __slots__ = ['args']
    def __init__(self, *args):
        self.args = args
    def to_ast(self) ->'ast.Tuple':
        return ASTUtils.ast_tuple(*self.args)
class AbstractSet(AbstractExpr):
    __tag__ = 'Set'
    __slots__ = ['args']
    def __init__(self, *args):
        self.args = args
    def to_ast(self) ->'ast.Set':
        return ASTUtils.ast_set(*self.args)
class AbstractDict(AbstractExpr):
    __tag__ = 'Dict'
    __slots__ = ['kwargs']
    def __init__(self, **kwargs):
        self.kwargs = kwargs
    def to_ast(self) ->'ast.Dict':
        return ASTUtils.ast_dict(**self.kwargs)

class AbstractListComp(AbstractExpr):
    __tag__ = 'ListComp'
    __slots__ = ['expr', 'var', 'iterable', 'filter']
    def __init__(self, expr, var, iterable, filter=None):
        self.expr, self.var, self.iterable, self.filter = expr, var, iterable, filter
    def to_ast(self) ->'ast.ListComp':
        return ASTUtils.ast_list_comprehension(self.expr, self.var, self.iterable, filter=self.filter)
class AbstractGenerator(AbstractExpr):
    __tag__ = 'Generator'
    __slots__ = ['expr', 'var', 'iterable', 'filter']
    def __init__(self, expr, var, iterable, filter=None):
        self.expr, self.var, self.iterable, self.filter = expr, var, iterable, filter
    def to_ast(self) ->'ast.GeneratorExp':
        return ASTUtils.ast_generator(self.expr, self.var, self.iterable, filter=self.filter)
class AbstractSetComp(AbstractExpr):
    __tag__ = 'SetComp'
    __slots__ = ['expr', 'var', 'iterable', 'filter']
    def __init__(self, expr, var, iterable, filter=None):
        self.expr, self.var, self.iterable, self.filter = expr, var, iterable, filter
    def to_ast(self) ->'ast.SetComp':
        return ASTUtils.ast_set_comprehension(self.expr, self.var, self.iterable, filter=self.filter)
class AbstractDictComp(AbstractExpr):
    __tag__ = 'DictComp'
    __slots__ = ['key_expr', 'value_expr', 'var', 'iterable', 'filter']
    def __init__(self, key_expr, value_expr, var, iterable, filter=None):
        self.key_expr, self.value_exp, self.var, self.iterable, self.filter = key_expr, value_expr, var, iterable, filter
    def to_ast(self) -> 'ast.DictComp':
        return ASTUtils.ast_dict_comprehension(self.key_expr, self.value_exp, self.var, self.iterable, filter=self.filter)

class AbstractArgSpec(AbstractExpr):
    __tag__ = 'ArgSpec'
    __slots__ = ['args', 'kwargs']
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
    def to_ast(self) ->'ast.arguments':
        return ASTUtils.ast_arglist((self.args, self.kwargs))
    def __repr__(self):
        return "{}({}, {})".format(
            type(self).__name__,
            self.args,
            self.kwargs
        )

class AbstractLambda(AbstractExpr):
    __tag__ = 'Lambda'
    __slots__ = ['spec', 'body']
    def __init__(self, *args, **kwargs):
        if len(args) == 1 and isinstance(args[0], AbstractArgSpec):
            self.spec = args[0]
        else:
            self.spec = AbstractArgSpec(*args, **kwargs)
        self.body = None
    def __call__(self, *args, **kwargs):
        if self.body is None:
            if len(kwargs) == 0:
                if len(args) > 1:
                    args = [AbstractList(*args)[-1]]
                self.body = args[0]
                return self
            raise ValueError("Lambda expression has no body yet...")
        return super().__call__(*args, **kwargs)
    def to_ast(self) ->'ast.Lambda':
        if self.body is None:
            raise ValueError("Lambda expression has no body yet...")
        return ASTUtils.ast_lambda(self.spec, self.body)
    def __repr__(self):
        return "{}({})({})".format(
            type(self).__name__,
            self.spec,
            self.body
        )

class AbstractIfExp(AbstractExpr):
    __tag__ = 'If'
    __slots__ = ['test', 'body', 'else_branch']
    def __init__(self, test, body, orelse=None):
        self.test, self.body, self.else_branch = test, body, orelse
    def orelse(self, orelse):
        self.else_branch = orelse
    def to_ast(self):
        if self.else_branch is None:
            raise ValueError("IfExp needs an else branch")
        return ASTUtils.ast_ternary(self.test, self.body, self.else_branch)

# def AbstractDecorator(abs_fun)

# TODO: Eventually I'll need proper module support (`for`, `def`, `class`) so I can "compile" programs against a stack

class Abstract:
    """
    Provides a namespace for the different abstract classes
    """

    @classmethod
    def vars(cls, *spec, symbol_type=None):
        if symbol_type is None:
            symbol_type = AbstractName
        if not isinstance(spec[0], str):
            spec = spec[0]
        if len(spec) == 1:
            spec = spec[0].split()
        return [symbol_type(x) for x in spec]

    Expr = AbstractExpr
    Name = AbstractName

    Lambda = AbstractLambda

    List = AbstractList
    Tuple = AbstractTuple
    Set = AbstractSet
    Dict = AbstractDict

    # Lambda = AbstractLambda

#     @classmethod
#     def name_redir(self, item):
#         return self.Name[item]
# Abstract.__getitem__ = Abstract.name_redir